# -*- coding: utf-8 -*-
"""
Lower-level and simpler NSID-specific HDF5 utilities that facilitate
higher-level data operations

Created on Tue Aug  3 21:14:25 2020

@author: Gerd Duscher, and Suhas Somnath
"""
from __future__ import division, print_function, absolute_import, unicode_literals
import sys
from warnings import warn
import h5py
import numpy as np
import datetime
import ase

from sidpy.hdf.hdf_utils import get_attr, copy_dataset, write_simple_attrs, \
    write_book_keeping_attrs, h5_group_to_dict
from sidpy.hdf import hdf_utils as hut
from sidpy import Dimension, Dataset

pynsid_version = 'scifireader-0.0.0'

if sys.version_info.major == 3:
    unicode = str


def get_all_main(parent, verbose=False):
    """
    Simple function to recursively print the contents of a hdf5 group
    Parameters
    ----------
    parent : :class:`h5py.Group`
        HDF5 Group to search within
    verbose : bool, optional. Default = False
        If true, extra print statements (usually for debugging) are enabled
    Returns
    -------
    main_list : list of h5py.Dataset
        The datasets found in the file that meet the 'Main Data' criteria.
    """
    if not isinstance(parent, (h5py.Group, h5py.File)):
        raise TypeError('parent should be a h5py.File or h5py.Group object')

    main_list = list()

    def __check(name, obj):
        if verbose:
            print(name, obj)
        if isinstance(obj, h5py.Dataset):
            if verbose:
                print(name, 'is an HDF5 Dataset.')
            ismain = check_if_main(obj)
            if ismain:
                if verbose:
                    print(name, 'is a `Main` dataset.')
                # TODO: Upconvert to sidpy.Dataset object with new function
                main_list.append(obj)

    if verbose:
        print('Checking the group {} for `Main` datasets.'.format(parent.name))
    parent.visititems(__check)

    return main_list


def read_h5py_dataset(dset):
    if not isinstance(dset, h5py.Dataset):
        raise TypeError('can only read single Dataset, use read_all_in_group or read_all function instead')

    if not check_if_main(dset):
        raise TypeError('can only read NSID datasets, not general one, try to import with from_array')

    # create vanilla dask array
    #find out if data is complex or not
    if 'complex' in str(dset.dtype): 
        data_array = np.empty(dset.shape, dtype = 'complex64')
    else:
        data_array = np.empty(dset.shape, dtype = 'float')

    dset.read_direct(data_array)
    dataset = Dataset.from_array(data_array)

    if 'title' in dset.attrs:
        dataset.title = dset.attrs['title']
    else:
        dataset.title = dset.name

    if 'units' in dset.attrs:
        dataset.units = dset.attrs['units']
    else:
        dataset.units = 'generic'

    if 'quantity' in dset.attrs:
        dataset.quantity = dset.attrs['quantity']
    else:
        dataset.quantity = 'generic'

    if 'data_type' in dset.attrs:
        dataset.data_type = dset.attrs['data_type']
    else:
        dataset.data_type = 'generic'

    if 'modality' in dset.attrs:
        dataset.modality = dset.attrs['modality']
    else:
        dataset.modality = 'generic'

    if 'source' in dset.attrs:
        dataset.source = dset.attrs['source']
    else:
        dataset.source = 'generic'

    dataset.axes = {}

    for dim in range(np.array(dset).ndim):
        try:
            label = dset.dims[dim].keys()[-1]
            name = dset.dims[dim][label].name
            dim_dict = {'quantity': 'generic', 'units': 'generic', 'dimension_type': 'generic'}
            dim_dict.update(dict(dset.parent[name].attrs))

            dataset.set_dimension(dim, Dimension(np.array(dset.parent[name][()]),
                                                 dset.dims[dim].label,
                                                 dim_dict['quantity'], dim_dict['units'],
                                                 dim_dict['dimension_type']))
        except ValueError:
            print('dimension {} not NSID type using generic'.format(dim))

    for key in dset.parent:
        if isinstance(dset.parent[key], h5py.Group):
            if 'Structure_' in key:
                structure_group = dset.parent[key]
                structure = h5_group_to_dict(structure_group)
                atoms, atoms_name = ase_from_dictionary(structure[key])

                dataset.structures.update({atoms_name: atoms})
            elif key[0] != '_':
                setattr(dataset, key, h5_group_to_dict(dset.parent[key])[key])
                
    dataset.h5_dataset = dset
    dataset.h5_filename = dset.file.filename
    try:
        dataset.h5_dataset_name = dset.name
    except ValueError:
        dataset.h5_dataset_name = ''
    return dataset


def ase_from_dictionary(tags):
    """
    converts structure dictionary to ase.Atoms object
    """
    atoms = None
    ase_name = ''
    for key, ase_dict in tags.items():
        if key not in ['machine_id', 'platform', 'sidpy_version', 'timestamp']:
            atoms = ase.Atoms(cell=ase_dict['unit_cell'],
                              symbols=ase_dict['elements'],
                              scaled_positions=ase_dict['base'])
            if 'info' in ase_dict:
                atoms.info = ase_dict['info']
            ase_name = key
    return atoms, ase_name


def find_dataset(h5_group, dset_name):
    """
    Uses visit() to find all datasets with the desired name
    Parameters
    ----------
    h5_group : :class:`h5py.Group`
        Group to search within for the Dataset
    dset_name : str
        Name of the dataset to search for
    Returns
    -------
    datasets : list
        List of [Name, object] pairs corresponding to datasets that match `ds_name`.
    """
    datasets = []

    for obj in hut.find_dataset(h5_group, dset_name):
        if check_if_main(obj):
            # TODO: Upconvert to sidpy.Dataset object with new function
            datasets.append(obj)
        else:
            datasets.append(obj)

    return datasets


def check_if_main(h5_main, verbose=False):
    """
    Checks the input dataset to see if it has all the necessary
    features to be considered a Main dataset.  This means it is
    dataset has dimensions of correct size and has the following attributes:
    * quantity
    * units
    * main_data_name
    * data_type
    * modality
    * source
    In addition, the shapes of the ancillary matrices should match with that of
    h5_main
    Parameters
    ----------
    h5_main : HDF5 Dataset
        Dataset of interest
    verbose : Boolean (Optional. Default = False)
        Whether or not to print statements
    Returns
    -------
    success : Boolean
        True if all tests pass
    """
    if not isinstance(h5_main, h5py.Dataset):
        if verbose:
            print('{} is not an HDF5 Dataset object.'.format(h5_main))
        return False

    number_of_dims = 0
    for dim in h5_main.dims:
        if np.array(dim.values()).size > 0:
            number_of_dims += 1

    if len(h5_main.shape) != number_of_dims:
        if verbose:
            print('Main data does not have full set of dimension scales. '
                  'Provided object has shape: {} but only {} dimensional '
                  'scales'.format(h5_main.shape, len(h5_main.dims)))
        return False

    # h5_name = h5_main.name.split('/')[-1]
    h5_group = h5_main.parent

    # success = True

    # Check for Datasets

    attrs_names = ['dimension_type', 'name', 'quantity', 'units']

    # Check for all required attributes in dataset
    main_attrs_names = ['quantity', 'units', 'main_data_name', 'pyNSID_version', 'data_type', 'modality', 'source']
    main_attr_success = np.all([att in h5_main.attrs for att in main_attrs_names])
    if verbose:
        print('All Attributes in dataset: ', main_attr_success)
    if not main_attr_success:
        if verbose:
            print('{} does not have the mandatory attributes'.format(h5_main.name))
        return False

    for attr_name in main_attrs_names:
        val = get_attr(h5_main, attr_name)
        if not isinstance(val, (str, unicode)):
            if verbose:
                print('Attribute {} of {} found to be {}. Expected a string'.format(attr_name, h5_main.name, val))
            return False

    length_success = []
    dset_success = []
    attr_success = []
    # Check for Validity of Dimensional Scales
    for i, dimension in enumerate(h5_main.dims):
        # check for all required attributes
        h5_dim_dset = h5_group[dimension.label]

        attr_success.append(np.all([att in h5_dim_dset.attrs for att in attrs_names]))
        dset_success.append(isinstance(h5_dim_dset, h5py.Dataset))
        # dimensional scale has to be 1D
        if len(h5_dim_dset.shape) == 1:
            # and of the same length as the shape of the dataset
            length_success.append(h5_main.shape[i] == h5_dim_dset.shape[0])
        else:
            length_success.append(False)
    # We have the list now and can get error messages according to which dataset is bad or not.
    if np.all([np.all(attr_success), np.all(length_success), np.all(dset_success)]):
        if verbose:
            print('Dimensions: All Attributes: ', np.all(attr_success))
            print('Dimensions: All Correct Length: ', np.all(length_success))
            print('Dimensions: All h5 Datasets: ', np.all(dset_success))
    else:
        if False in length_success:
            print('length of dimension scale {} is wrong'.format(length_success.index(False)))
        if False in attr_success:
            print('attributes in dimension scale {} are wrong'.format(attr_success.index(False)))
        if False in dset_success:
            print('dimension scale {} is not a dataset'.format(dset_success.index(False)))
        return False

    return main_attr_success


def link_as_main(h5_main, dim_dict):
    """
    Attaches datasets as h5 Dimensional Scales to  `h5_main`

    Parameters
    ----------
    h5_main : h5py.Dataset
        N-dimensional Dataset which will have the references added as h5 Dimensional Scales
    dim_dict: dictionary with dimensional order as key and items are datasets to be used as h5 Dimensional Scales

    Returns
    -------
    pyNSID.NSIDataset
        NSIDataset version of h5_main now that it is a NSID Main dataset
    """
    if not isinstance(h5_main, h5py.Dataset):
        raise TypeError('h5_main should be a h5py.Dataset object')
    if not isinstance(dim_dict, dict):
        raise TypeError("""dim_dict must be a dictionary"""
                        """ (keys: dimensional order, values: h5py.Datasets)""")

    h5_parent_group = h5_main.parent
    main_shape = h5_main.shape
    ######################
    # Validate Dimensions
    ######################
    # An N dimensional dataset should have N items in the dimension dictionary
    if len(dim_dict) != len(main_shape):
        raise ValueError('Incorrect number of dimensions: {} provided to support main data, of shape: {}'
                         .format(len(dim_dict), main_shape))
    if set(range(len(main_shape))) != set(dim_dict.keys()):
        raise KeyError('')
    
    dim_names = []
    for index, dim_exp_size in enumerate(main_shape):
        this_dim = dim_dict[index]
        if isinstance(this_dim, h5py.Dataset):
            error_message = validate_h5_dimension(this_dim, main_shape[index])
            if len(error_message) > 0:
                raise TypeError('Dimension {} has the following error_message:\n'.format(index), error_message)
            else:
                # if h5_dim.name not in dim_names:
                if this_dim.name not in dim_names:  # names must be unique
                    dim_names.append(this_dim.name)
                else:
                    raise TypeError('All dimension names must be unique, found {} twice'.format(this_dim.name))
                if this_dim.file != h5_parent_group.file:
                    copy_dataset(this_dim, h5_parent_group, verbose=False)
        else: 
            raise TypeError('Items in dictionary must all  be h5py.Datasets !')

    ################
    # Attach Scales
    ################
    for i, this_dim_dset in dim_dict.items():
        this_dim_dset.make_scale(this_dim_dset.attrs['name'])
        h5_main.dims[int(i)].label = this_dim_dset.attrs['name']
        h5_main.dims[int(i)].attach_scale(this_dim_dset)
        
    return h5_main


def validate_h5_dimension(h5_dim, dim_length):
    """
    Validates a dimension already present in an HDF5 file.

    Parameters
    ----------
    h5_dim : h5py.Dataset
        HDF5 dataset which represents a scientific dimension.
        The dimension should have non-empty attributes 'name', quantity',
        'units', and 'dimension_type'
    dim_length : int
        Expected length of dataset

    Returns
    -------
    error_message: string, empty if ok.

    Notes
    -----
    A valid dataset to be used as dimension must be 1D not a compound data type but 'simple'.
    Such a dataset must have  ancillary attributes 'name', quantity', 'units', and 'dimension_type',
    which have to be of  types str, str, str, and bool respectively and not empty
    If it is not valid of dataset, Exceptions are raised.
    """
    warn('validate_h5_dimension may be removed in a future version',
         FutureWarning)
    # TODO: Raise exceptions instead of returning strings that need to be parsed

    if not isinstance(h5_dim, h5py.Dataset):
        error_message = 'this Dimension must be a h5 Dataset'
        return error_message

    error_message = ''
    # Is it 1D?
    if len(h5_dim.shape) != 1:
        error_message += ' High dimensional datasets are not allowed as dimensions;\n'
    # Does this dataset have a "simple" dtype - no compound data type allowed!
    # is the shape matching with the main dataset?
    if len(h5_dim) != dim_length:
        error_message += ' Dimension has wrong length;\n'
    # TODO: Relax requirements for these attributes. Check against sidpy.Dataset
    # Does it contain some ancillary attributes like 'name', quantity', 'units', and 'dimension_type'
    necessary_attributes = ['name', 'quantity', 'units', 'dimension_type']
    for key in necessary_attributes:
        if key not in h5_dim.attrs:
            error_message += 'Missing {} attribute in dimension;\n '.format(key)
        elif not isinstance(h5_dim.attrs[key], str):
            error_message += '{} attribute in dimension should be string;\n '.format(key)

    return error_message


def validate_main_and_dims(main_shape, dim_dict, h5_parent_group):
    """
    Validates the shape of the main dataset against the dictionary of
    dimensions and the parent HDF5 group into which the data would be written.
    This method was written as a low-level validation check before
    sidpy.Dataset was conceived. It may still be relevant if one intends to
    reuse the Dimension HDF5 datasets already in the file

    Parameters
    ----------
    main_shape : list or tuple
        Shape of the Main dataset
    dim_dict : dict
        Dictionary of dimensions that could either be sidpy.Dimension or
        h5py.Dataset objects
    h5_parent_group : h5py.Group
        HDF5 group to write into

    Returns
    -------
    bool
        Whether or not the dimensions match the main data shape
    """
    warn('validate_main_and_dims may not exist in a future version',
         FutureWarning)

    # Each item could either be a Dimension object or a HDF5 dataset
    # Collect the file within which these ancillary HDF5 objects are present if they are provided
    which_h5_file = {}
    # Also collect the names of the dimensions. We want them to be unique
    dim_names = []

    dimensions_correct = []
    for index, dim_exp_size in enumerate(main_shape):
        this_dim = dim_dict[index]
        if isinstance(this_dim, h5py.Dataset):
            error_message = validate_h5_dimension(this_dim, main_shape[index])

            # All these checks should live in a helper function for cleanliness

            if len(error_message) > 0:
                print('Dimension {} has the following error_message:\n'.format(index), error_message)

            else:
                if this_dim.name not in dim_names:  # names must be unique
                    dim_names.append(this_dim.name)
                else:
                    raise TypeError('All dimension names must be unique, found'
                                    ' {} twice'.format(this_dim.name))

                # are all datasets in the same file?
                if this_dim.file != h5_parent_group.file:
                    copy_dataset(this_dim, h5_parent_group, verbose=True)

        elif isinstance(this_dim, Dimension):
            # is the shape matching with the main dataset?
            dimensions_correct.append(len(this_dim.values) == dim_exp_size)
            # Is there a HDF5 dataset with the same name already in the provided group
            # where this dataset will be created?
            if this_dim.name in h5_parent_group:
                # check if this object with the same name is a dataset and if it satisfies the above tests
                if isinstance(h5_parent_group[this_dim.name], h5py.Dataset):
                    print('needs more checking')
                    dimensions_correct[-1] = False
                else:
                    dimensions_correct[-1] = True
            # Otherwise, just append the dimension name for the uniqueness test
            elif this_dim.name not in dim_names:
                dim_names.append(this_dim.name)
            else:
                dimensions_correct[-1] = False
        else:
            raise TypeError('Values of dim_dict should either be h5py.Dataset '
                            'objects or Dimension. Object at index: {} was of '
                            'type: {}'.format(index, index))

        for dim in which_h5_file:
            if which_h5_file[dim] != h5_parent_group.file.filename:
                print('need to copy dimension', dim)
        for i, dim_name in enumerate(dim_names[:-1]):
            if dim_name in dim_names[i + 1:]:
                print(dim_name, ' is not unique')

    return all(dimensions_correct)


def write_pynsid_book_keeping_attrs(h5_object):
    """
    Writes bookkeeping information to the HDF5 object

    Parameters
    ----------
    h5_object

    Returns
    -------

    """
    write_book_keeping_attrs(h5_object)
    write_simple_attrs(h5_object, {'pyNSID_version': pynsid_version})


def make_nexus_compatible(h5_dataset):
    """
    Makes a pyNSID file compatible with the NeXus file format
    by adding the appropriate attributes and by writing one group.

    Parameters
    ----------
    h5_dataset: h5py.Dataset
        h5py dataset with main data

    Returns
    -------

    """
    if not isinstance(h5_dataset, h5py.Dataset):
        raise ValueError('We need a h5py dataset for compatibility with NeXus file format')

    time_stamp = "T".join(str(datetime.datetime.now()).split())
    h5_file = h5_dataset.file
    h5_group = h5_dataset.parent.parent
    h5_file.attrs[u'default'] = h5_dataset.parent.parent.name

    # give the HDF5 root some more attributes

    h5_file.attrs[u'file_name'] = h5_file.filename
    h5_file.attrs[u'file_time'] = time_stamp
    h5_file.attrs[u'instrument'] = u'None'
    h5_file.attrs[u'creator'] = u'pyNSID'
    h5_file.attrs[u'NeXus_version'] = u'4.3.0'
    h5_file.attrs[u'HDF5_Version'] = h5py.version.hdf5_version
    h5_file.attrs[u'h5py_version'] = h5py.version.version

    h5_file.attrs[u'default'] = h5_group.name

    h5_group.attrs[u'NX_class'] = u'NXentry'
    h5_group.attrs[u'default'] = h5_dataset.name.split('/')[-1]

    if 'title' in h5_group:
        del h5_group['title']
    h5_group.create_dataset(u'title', data=h5_dataset.name.split('/')[-1])

    nxdata = h5_dataset.parent

    nxdata.attrs[u'NX_class'] = u'NXdata'
    nxdata.attrs[u'signal'] = h5_dataset.name.split('/')[-1]
    axes = []
    for dimension_label in h5_dataset.attrs['DIMENSION_LABELS']:
        axes.append(dimension_label.decode('utf8'))

    nxdata.attrs[u'axes'] = axes
    for i, axis in enumerate(axes):
        nxdata.attrs[axes[i] + '_indices'] = [i, ]


# -*- coding: utf-8 -*-

#This reader comes from NanoSurf (https://github.com/syrupface/NSFopen/blob/master/NSFopen/read/read.py)

"""
Created on Mon Apr 23 15:17:03 2018

Copyright (c) 2018 nelson<at>nanosurf.com

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

"""


import re
import numpy as np
import h5py
import time
import pandas as pd
import sidpy as sid
from sidpy.sid import Reader

class nid_read():
    def __init__(self, filename=None, dataframe=True, verbose=True):
        self.filename = filename
        self.__dataframe = dataframe
        self.verbose = verbose
        # names of the channels that can be opened.  Can be added to.
        self.data_names = ['Spec forward', 'Spec backward', 'Spec fwd pause', 'Spec bwd pause',
                           'Scan forward', 'Scan backward', '2nd scan forward', '2nd scan backward',
                           'Frequency sweep', 'Frequency sweep SHO', 'Spectrum FFT', 'Spectrum Fit']
        self.data_types = [1, 1, 1, 1, 0, 0, 0, 0, 2, 2, 2, 2]
        self.data = None
        self.param = None
        if not filename:
            print('\nFile required!\n')
            return
        self.read()

    def read(self):
        t = time.time()

        encode = 'ANSI'
        # encode = 'UTF-8'
        encode = 'ISO-8859-1'

        # isolates header from data file
        if self.verbose:
            print("Reading Header")

        fid = open(self.filename, 'rb')
        data_in = fid.read().split(b'\n\r\n\r')
        fid.close()

        head = np.array(data_in[0].split(b'\r\n'))
        boo = head == b''
        indices = np.nonzero(boo[1:] != boo[:-1])[0] + 1
        head = np.split(head, indices)

        head = [h for h in head if len(h) > 1]

        header = {}
        for h in head:
            title = h[0].decode(encode)[1:-1]
            data = h[1:]
            values = {}
            for d in data:
                val = d.decode(encode).split('=')
                values[val[0]] = val[1].rstrip()
            header[title] = values

        param = {}

        def from_array(key, param):
            return float(header[key][param].split(',')[4])

        def get_values(key, param):
            string = header[key][param]

            ty, val, unit = re.search(r'(\S)\[(.*)\]\*\[(.*)\]', string).groups()
            if ty == 'D':
                val = float(val)
            if ty == 'B':
                val = bool(val)
            if ty == 'L':
                val = int(val)
            if ty == 'V':
                val = [float(v) for v in val.split(',')]
                unit = [str(u) for u in unit.split(',')]
            return {'Value': val, 'Unit': unit}

        groupNames = []
        dataSet = []

        ds = 'DataSet'
        grCnt = int(header[ds]['GroupCount'])
        for gr in range(grCnt):
            groupNames.append(header[ds]['Gr' + str(gr) + '-Name'])
            gc = int(header[ds]['Gr' + str(gr) + '-Count'])

            for g in range(gc):
                key = 'Gr' + str(gr) + '-Ch' + str(g)
                if key in header[ds]:
                    dataSet.append(header[ds][key])

        terms = ['Frame', 'Points', 'Lines', 'SaveBits',
                 'Dim0Min', 'Dim0Range', 'Dim0Unit',
                 'Dim1Min', 'Dim1Range', 'Dim1Unit',
                 'Dim2Min', 'Dim2Range', 'Dim2Unit',
                 'Dim2Name']

        types = [0, 1, 1, 1,
                 2, 2, 0,
                 2, 2, 0,
                 2, 2, 0,
                 0]

        # 0 = string, 1 = integer, 2 = float
        typ = [lambda x:str(x), lambda x:int(x), lambda x:float(x)]
     
        for term, ty in zip(terms, types):
            val = []
            for d in dataSet:
                h = header[d]
                if term in h:
                    val.append(typ[ty](h[term]))

            param[term] = val

        # for spectroscopy with maps
        dataPoints = []
        for d in dataSet:
            h = header[d]
            linePoints = []
            if 'LineDim0Min' in h:
                for key, value in h.items():
                    if re.match(r'LineDim\d*Points', key):
                        linePoints.append(int(value))
            dataPoints.append(np.array(linePoints))
        param['LinePoints'] = dataPoints
        
        cantilever = {}
        key = 'DataSet\\Calibration\\Cantilever'
        if key in header:
            cantilever = {'Manufacturer': header[key]['Manufacturer'], 'Name': header[key]['Name']}

            propcount = int(header[key]['PropCount'])
            for i in range(propcount):
                cantilever['Prop' + str(i)] = get_values(key, 'Prop' + str(i))

        key = 'DataSet\\Calibration\\Scanhead'
        if key in header:
            sens = from_array(key, 'In5')
            cantilever['Sensitivity'] = {'Value': sens / 10, 'Unit': 'm'}

        mapTable = {}  
        key = 'DataSet\\SpecInfos'
        if key in header:
            n = int(header[key]['SubSectionCount'])
            secNames = [header[key]['SubSection' + str(i)] for i in range(n)]
            
            specMode = header[key + '\\' + secNames[0]]['SpecMode']
            
            count = int(header[key + '\\' + secNames[1]]['Count'])
            
            mapTable = [header[key + '\\' + secNames[1]][specMode[:3] + str(i)].split(';') for i in range(count)]
            mapTable = np.array(mapTable).astype(np.float)
            
        scanOffset = {}
        key = 'DataSet\\Parameters\\Imaging'
        if key in header:
            scanOffset = get_values(key, 'ScanOffset')
        
        thermal = {}
        key = 'DataSet-Info'
        if '-- Thermal Tuning --' in header[key]:
            if 'k' in header[key]['Q Factor:']:
                Q = float(header[key]['Q Factor:'][:-1])*1e3
            else:
                Q = float(header[key]['Q Factor:'])
            Freq = header[key]['Frequency:']
            SpringK = header[key]['Spring Constant:']
            Peak = header[key]['Peak Value:']
            
            gr = re.search(r'(\d+\.\d+)(\S*)', Freq).groups()
            Freq = {'Value': gr[0], 'Unit': gr[1]}
            try:
                gr = re.search(r'(\d+\.\d+)\s(\S*)', SpringK).groups()
            except:
               gr = re.search(r'(\d)\s(\S*)', SpringK).groups()
            SpringK = {'Value': gr[0], 'Unit': gr[1]}
            
            try:
                gr = re.search(r'(\d+\.\d+)(\S*)', Peak).groups()
            except:
                gr = re.search(r'(\d)(\S*)', Peak).groups()
            Peak = {'Value': gr[0], 'Unit': gr[1]}
            
            thermal = {'Frequency': Freq, 'Q Factor': Q, 'Spring Constant': SpringK, 'Peak': Peak}
        
        required_size = int(sum([a*b*c/8 for a, b, c in zip(param['Points'],
                                                            param['Lines'],
                                                            param['SaveBits'])]))
        
        q = float(2**param['SaveBits'][0])
        z0 = float(q/2)

        if param['SaveBits'][0] == 16:
            dt = np.int16
        elif param['SaveBits'][0] == 32:
            dt = np.int32
        
        if self.verbose:
            print('Reading Data')
        
        fid = open(self.filename, 'rb')
        fid.seek(-required_size, 2)
        data_in = np.fromfile(fid, count=required_size, dtype=dt).astype(float)
        data_in = (data_in + z0) / q
        fid.close()

        data_in = np.split(data_in, np.cumsum([a * b for a, b in zip(param['Points'],
                                                                     param['Lines'])]))
        data_in.pop()

        data = []
        # reshape and rescale data
        for datain, pts, lns, zran, zmin in zip(data_in,
                                                param['Points'],
                                                param['Lines'],
                                                param['Dim2Range'],
                                                param['Dim2Min']):

            data.append(datain.reshape((lns, pts)).__mul__(zran).__add__(zmin))

        data_crop = []
        for numP, dSet in zip(param['LinePoints'], data):
            if numP.any():
                temp = []
                for n, d in zip(numP, dSet):
                    temp.append(d[:n])
                data_crop.append(temp)
            else:
                data_crop.append(dSet)
        data = np.array(data_crop, dtype=float)

        # this is the index of the channels that will be output.
        idx = [idx for idx, frame in enumerate(param['Frame']) if frame in self.data_names]

        # only take data from approved list in name
        data = data[idx]
        frames = np.array(param['Frame'])[idx]
        channel = np.array(param['Dim2Name'])[idx]

        out = {}
        for frame, chan, dat in zip(frames, channel, data):
            if frame not in out:
                out[frame] = {}
            if chan not in out[frame]:
                out[frame][chan] = {}
            out[frame][chan] = dat
    
        image = {}
        spec = {}
        spectrum = {}

        for name, value in out.items():
            dt = self.data_types[self.data_names.index(name)]
            if dt == 0:
                if name == 'Scan forward':
                    image['Forward'] = value
                elif name == 'Scan backward':
                    image['Backward'] = value
                elif name == '2nd scan forward':
                    image['2nd Forward'] = value
                elif name == '2nd scan backward':
                    image['2nd Backward'] = value
                else:
                    pass
            if dt == 1:
                if name == 'Spec forward':
                    spec['Forward'] = value
                elif name == 'Spec backward':
                    spec['Backward'] = value
                elif name == 'Spec fwd pause':
                    spec['Pause Forward'] = value
                elif name == 'Spec bwd pause':
                    spec['Pause Backward'] = value
                else:
                    pass
            if dt == 2:
                if name == 'Spectrum FFT':
                    spectrum['FFT'] = value
                elif name == 'Spectrum Fit':
                    spectrum['Fit'] = value
                elif name == 'Frequency sweep':
                    spectrum['Sweep'] = value
                elif name == 'Frequency sweep SHO':
                    spectrum['SHO'] = value
                else:
                    pass

        x = dict(zip(['min', 'range', 'units'],
                     [param['Dim0Min'], param['Dim0Range'], param['Dim0Unit']]))
        y = dict(zip(['min', 'range', 'units'],
                     [param['Dim1Min'], param['Dim1Range'], param['Dim1Unit']]))
        z = dict(zip(['min', 'range', 'units'],
                     [param['Dim2Min'], param['Dim2Range'], param['Dim2Unit']]))
        specMap = dict(zip(['maps'], [mapTable]))
        offset = dict(zip(['offset'], [scanOffset]))

        parameters = dict(zip(['Tip', 'X', 'Y', 'Z', 'SpecMap', 'OffSet', 'Tune', 'Header Dump'],
                              [cantilever, x, y, z, specMap, offset, thermal, header]))

        dataout = [image, spec, spectrum]
        datanames = ['Image', 'Spec', 'Sweep']

        valid_data = [val for num, val in enumerate(dataout) if val]
        valid_names = [datanames[num] for num, val in enumerate(dataout) if val]

        dataout = dict(zip(valid_names, valid_data))
        t_end = time.time() - t

        if self.verbose:
            print('Elapsed Time: %3.2f sec\n' % t_end)

        if self.__dataframe:
            # import pandas as pd
            self.data_pd = self.__toPandaDF(dataout).unstack(level=0)
            self.data = dataout#self.__toPandaDF(dataout).unstack(level=0)
            # self.specunits = pd.DataFrame(specUnits)
            self.param = self.__toPandaSeries(parameters)
        else:
            self.data = dataout
            self.param = parameters

    @staticmethod
    def __toPandaDF(user_dict):
        import pandas as pd
        df = pd.DataFrame.from_dict({(i, j): user_dict[i][j]
                                     for i in user_dict.keys()
                                     for j in user_dict[i].keys()}, orient='index').transpose()
        return df

    @staticmethod
    def __toPandaSeries(user_dict):
        import pandas as pd
        df = pd.Series({(i, j): user_dict[i][j]
                        for i in user_dict.keys()
                        for j in user_dict[i].keys()})
        return df

class nhf_read():
    def __init__(self, filename=None):
        self.filename = filename
        self.fopen = h5py.File(filename, 'r')
        self.read()

    def read(self):
# measurement
        group_name = list(self.fopen.keys())
        # forward/backward
        segment = self.fopen[group_name[0]]
        seg_names = list(segment.keys())

        s_prop = segment.attrs
        prop_s = {name:s_prop[name] for name in list(s_prop)}

        DAC_max = (2**31-1)

        data_out = {}
        prop = []
        
        for s in seg_names:
            meas = segment[s]
            direction = meas.attrs['name']
            data_out[direction] = {}
            
            for m in list(meas.keys()):
                data = meas[m]
                d_prop = data.attrs
                
                Nx = prop_s['image_points_per_line']
                Ny = prop_s['image_number_of_lines']
                
                prop.append({name:d_prop[name] for name in list(d_prop)})
                
                bit = d_prop['dataset_size']
                mx = d_prop['value_max']
                mn= d_prop['value_min']
                cal_max = d_prop['base_calibration_max']
                cal_min = d_prop['base_calibration_min']
                
                channel = d_prop['name']
                scale = (cal_max-cal_min)/DAC_max/2
                
                data_out[direction][channel] = np.reshape(data[:], [Nx, Ny])*scale
        
        

        self.data_pd = pd.DataFrame(data_out).unstack(level=0)
        self.prop = prop_s
        self.data = data_out

  
class NanoSurfNIDReader(Reader):
    """
    Reads files obtained via NanoSurf controllers in .NID files.

    """

    def read(self, verbose=False):
        """
        Reads the file given in file_path into a sidpy dataset

        Parameters
        ----------
        verbose : Boolean (Optional)
            Whether or not to show  print statements for debugging

        Returns
        -------
        sidpy.Dataset : List of sidpy.Dataset objects.
            Multi-channel inputs are separated into individual dataset objects
        """

        file_path = self._input_file_path
        folder_path, file_name = path.split(file_path)
        
        # Extracting the raw data into memory
        reader = nid_read(filename=file_path)
        data_raw = reader.data
        parameters = reader.param.to_dict()
        channel_units = parameters[('Z', 'units')]
        datasets = {}

        for ind_l1, key_l1 in enumerate(data_raw['Image'].keys()):
            for ind_l2, key_l2 in enumerate(data_raw['Image'][key_l1].keys()):

                raw_data = data_raw['Image'][key_l1][key_l2]
                data_set = sid.Dataset.from_array(raw_data[:], name=key_l2)

                data_set.data_type = 'Image'
                data_set.units = channel_units[ind_l1]
                data_set.quantity = key_l2


                x_min = parameters[('X', 'min')][ind_l1]
                x_max = parameters[('X', 'range')][ind_l1]
                x_vec = np.linspace(x_min, x_max, data_set.shape[0])

                y_min = parameters[('Y', 'min')][ind_l1]
                y_max = parameters[('Y', 'range')][ind_l1]
                y_vec = np.linspace(y_min, y_max, data_set.shape[1])

                # Add dimension info
                data_set.set_dimension(0, sid.Dimension(x_vec, name='X',
                                                        units=channel_units[ind_l1], quantity='X',
                                                        dimension_type=sid.DimensionType.SPATIAL))

                data_set.set_dimension(1, sid.Dimension(y_vec, name='Y',
                                                        units=channel_units[ind_l1], quantity='Y',
                                                        dimension_type=sid.DimensionType.SPATIAL))

                # append metadata
                data_set.original_metadata = parameters
                key_name = key_l1 + " " + key_l2
                datasets[key_name] = data_set.copy()
               
        return datasets

   
    def can_read(self):
        """
        Tests whether or not the provided file has a .asc extension
        Returns
        -------

        """

        return super(NanoSurfNIDReader, self).can_read(extension='nid')



   


#!/usr/bin/env python
# -*- coding: iso-8859-1 -*-
"""
Will move to Scope Reader

################################################################################
# Python class for reading GATAN DM3 (DigitalMicrograph) files
# and extracting all metadata
# --
# tested on EELS spectra, spectrum images and single-image files and image-stacks
# --
# based on the DM3_Reader plug-in (v 1.3.4) for ImageJ by Greg Jefferis <jefferis@stanford.edu>
# http://rsb.info.nih.gov/ij/plugins/DM3_Reader.html
# --
# Python adaptation: Pierre-Ivan Raynal <raynal@med.univ-tours.fr>
# http://microscopies.med.univ-tours.fr/
#
# Extended for EELS support by Gerd Duscher, UTK 2012
# Rewritten for integration of sidpy 2020
#
# Works for python 3
#
################################################################################
"""

from __future__ import division, print_function, absolute_import, unicode_literals

import struct
import time
import numpy

import sys
import numpy as np
import os

import sidpy

version = '0.1beta'

debugLevel = 0  # 0=none, 1-3=basic, 4-5=simple, 6-10 verbose

if sys.version_info.major == 3:
    unicode = str


# ### utility functions ###

# ## binary data reading functions ###


def read_long(f):
    """Read 4 bytes as integer in file f"""
    read_bytes = f.read(4)
    return struct.unpack('>l', read_bytes)[0]


def read_64_long(f):
    read_bytes = f.read(8)
    return struct.unpack('>Q', read_bytes)[0]


def read_short(f):
    """Read 2 bytes as integer in file f"""
    read_bytes = f.read(2)
    return struct.unpack('>h', read_bytes)[0]


def read_byte(f):
    """Read 1 byte as integer in file f"""
    read_bytes = f.read(1)
    return struct.unpack('>b', read_bytes)[0]


def read_bool(f):
    """Read 1 byte as boolean in file f"""
    read_val = read_byte(f)
    return read_val != 0


def read_char(f):
    """Read 1 byte as char in file f"""
    read_bytes = f.read(1)
    return struct.unpack('c', read_bytes)[0]


def read_string(f, length=1):
    """Read len bytes as a string in file f"""
    read_bytes = f.read(length)
    str_fmt = '>' + str(length) + 's'
    return struct.unpack(str_fmt, read_bytes)[0]


def read_le_short(f):
    """Read 2 bytes as *little endian* integer in file f"""
    read_bytes = f.read(2)
    return struct.unpack('<h', read_bytes)[0]


def read_le_long(f):
    """Read 4 bytes as *little endian* integer in file f"""
    read_bytes = f.read(4)
    return struct.unpack('<l', read_bytes)[0]


def read_leu_short(f):
    """Read 2 bytes as *little endian* unsigned integer in file f"""
    read_bytes = f.read(2)
    return struct.unpack('<H', read_bytes)[0]


def read_leu_long(f):
    """Read 4 bytes as *little endian* unsigned integer in file f"""
    read_bytes = f.read(4)
    return struct.unpack('<L', read_bytes)[0]


def read_le_float(f):
    """Read 4 bytes as *little endian* float in file f"""
    read_bytes = f.read(4)
    return struct.unpack('<f', read_bytes)[0]


def read_le_double(f):
    """Read 8 bytes as *little endian* double in file f"""
    read_bytes = f.read(8)
    return struct.unpack('<d', read_bytes)[0]


# constants for encoded data types ##
SHORT = 2
LONG = 3
USHORT = 4
ULONG = 5
FLOAT = 6
DOUBLE = 7
BOOLEAN = 8
CHAR = 9
OCTET = 10
STRUCT = 15
STRING = 18
ARRAY = 20

# - association data type <--> reading function
readFunc = {
    SHORT: read_le_short,
    LONG: read_le_long,
    USHORT: read_leu_short,
    ULONG: read_leu_long,
    FLOAT: read_le_float,
    DOUBLE: read_le_double,
    BOOLEAN: read_bool,
    CHAR: read_char,
    OCTET: read_char,  # difference with char???
}

# other constants ##
IMGLIST = "root.ImageList."
OBJLIST = "root.DocumentObjectList."
MAXDEPTH = 64


# END constants ##

class DM3Reader(sidpy.Reader):
    debugLevel = -1

    """
    file_path: filepath to dm3 file.

    warn('This Reader will eventually be moved to the ScopeReaders package'
         '. Be prepared to change your import statements',
         FutureWarning)
    """

    def __init__(self, file_path, verbose=False):
        super().__init__(file_path)

        # initialize variables ##
        self.verbose = verbose
        self.__filename = file_path
        self.__chosen_image = -1

        # - open file for reading
        try:
            self.__f = open(self.__filename, 'rb')
        except FileNotFoundError:
            raise FileNotFoundError('File not found')

        # - create Tags repositories
        self.__stored_tags = {'DM': {}}

        # check if this is valid DM3 file
        is_dm = True
        # read header (first 3 4-byte int)
        # get version
        file_version = read_long(self.__f)

        # get indicated file size
        if file_version == 3:
            file_size = read_long(self.__f)
        elif file_version == 4:
            file_size = read_64_long(self.__f)
        else:
            is_dm = False
        # get byte-ordering
        le = read_long(self.__f)
        little_endian = (le == 1)

        if little_endian:
            self.endian_str = '>'
        else:
            self.endian_str = '<'

        # check file header, raise Exception if not DM3
        if not is_dm:
            raise TypeError("%s does not appear to be a DM3 or DM4 file." % os.path.split(self.__filename)[1])
        elif self.verbose:
            print("%s appears to be a DM3 file" % self.__filename)
        self.file_version = file_version
        self.file_size = file_size

        if self.verbose:
            print("Header info.:")
            print("- file version:", file_version)
            print("- le:", le)
            print("- file size:", file_size, "bytes")

        # don't read but close file
        self.__f.close()

    def read(self):
        try:
            self.__f = open(self.__filename, 'rb')
        except FileNotFoundError:
            raise FileNotFoundError('File not found')

        t1 = time.time()
        file_version = read_long(self.__f)

        # get indicated file size
        if file_version == 3:
            file_size = read_long(self.__f)
        elif file_version == 4:
            file_size = read_64_long(self.__f)
        else:
            is_dm = False
        # get byte-ordering
        le = read_long(self.__f)
        little_endian = (le == 1)

        if little_endian:
            self.endian_str = '>'
        else:
            self.endian_str = '<'

        # ... then read it

        self.__stored_tags = {'DM': {'file_version': file_version, 'file_size': file_size}}

        self.__read_tag_group(self.__stored_tags)

        if self.verbose:
            print("-- %s Tags read --" % len(self.__stored_tags))

        if self.verbose:
            t2 = time.time()
            print("| parse DM3 file: %.3g s" % (t2 - t1))

        path, file_name = os.path.split(self.__filename)
        basename, extension = os.path.splitext(file_name)
        dataset = sidpy.Dataset.from_array(self.data_cube, name=basename)
        self.__stored_tags['DM']['chosen_image'] = self.__chosen_image
        dataset.original_metadata = self.get_tags()

        self.set_dimensions(dataset)
        self.set_data_type(dataset)
        # convert linescan to spectral image
        if self.spectral_dim and dataset.ndim == 2:
            old_dataset = dataset.copy()
            meta = dataset.original_metadata.copy()
            basename = dataset.name
            data = np.array(dataset).reshape(dataset.shape[0], 1, dataset.shape[1])
            dataset = sidpy.Dataset.from_array(data, name=basename)
            dataset.original_metadata = meta
            dataset.set_dimension(0, old_dataset.dim_0)

            dataset.set_dimension(1, sidpy.Dimension([1], name='y', units='pixels',
                                                     quantity='distance', dimension_type='spatial'))
            dataset.set_dimension(2, old_dataset.dim_1)
            dataset.data_type = sidpy.DataType.SPECTRAL_IMAGE  # 'linescan'


        dataset.quantity = 'intensity'
        dataset.units = 'counts'
        dataset.title = basename
        dataset.modality = 'generic'
        dataset.source = 'DM3Reader'
        dataset.original_metadata['DM']['full_file_name'] = self.__filename

        return dataset

    def set_data_type(self, dataset):
        spectral_dim = False
        # print(dataset._axes)
        for dim, axis in dataset._axes.items():
            if axis.dimension_type == sidpy.DimensionType.SPECTRAL:
                spectral_dim = True
        self.spectral_dim = spectral_dim

        dataset.data_type = 'unknown'
        if 'ImageTags' in dataset.original_metadata['ImageList'][str(self.__chosen_image)]:
            image_tags = dataset.original_metadata['ImageList'][str(self.__chosen_image)]['ImageTags']
            if 'SI' in image_tags:
                if len(dataset.shape) == 3:
                    dataset.data_type = sidpy.DataType.SPECTRAL_IMAGE
                else:
                    if spectral_dim:
                        dataset.data_type = sidpy.DataType.SPECTRAL_IMAGE  # 'linescan'
                    else:
                        dataset.data_type = sidpy.DataType.IMAGE
                        dataset.metadata['image_type'] = 'survey image'

        if dataset.data_type == sidpy.DataType.UNKNOWN:
            if len(dataset.shape) > 3:
                raise NotImplementedError('Data_type not implemented yet')
            elif len(dataset.shape) == 3:
                if spectral_dim:
                    dataset.data_type = sidpy.DataType.SPECTRAL_IMAGE
                else:
                    dataset.data_type = 'image_stack'
            elif len(dataset.shape) == 2:
                if spectral_dim:
                    basename = dataset.name
                    dataset.data_type = sidpy.DataType.SPECTRAL_IMAGE
                else:
                    dataset.data_type = 'image'
            elif len(dataset.shape) == 1:
                if spectral_dim:
                    dataset.data_type = sidpy.DataType.SPECTRUM
                else:
                    dataset.data_type = sidpy.DataType.LINE_PLOT

    def set_dimensions(self, dataset):
        dimensions_dict = dataset.original_metadata['ImageList'][str(self.__chosen_image)]['ImageData']['Calibrations'][
            'Dimension']

        reciprocal_name = 'u'
        spatial_name = 'x'

        for dim, dimension_tags in dimensions_dict.items():
            # Fix annoying scale of spectrum_images in Zeiss  and SEM images
            if dimension_tags['Units'] == 'µm':
                dimension_tags['Units'] = 'nm'
                dimension_tags['Scale'] *= 1000.0

            if dimension_tags['Units'].strip() == '':
                units = 'counts'
            else:
                units = dimension_tags['Units']

            values = (np.arange(dataset.shape[int(dim)]) - dimension_tags['Origin']) * dimension_tags['Scale']

            if 'eV' == units:
                dataset.set_dimension(int(dim), sidpy.Dimension(values, name='energy_loss', units=units,
                                                                quantity='energy-loss',
                                                                dimension_type=sidpy.DimensionType.SPECTRAL))
            elif 'eV' in units:
                dataset.set_dimension(int(dim), sidpy.Dimension(values, name='energy', units=units,
                                                                quantity='energy',
                                                                dimension_type=sidpy.DimensionType.SPECTRAL))
            elif '1/' in units or units in ['mrad', 'rad']:
                dataset.set_dimension(int(dim), sidpy.Dimension(values, name=reciprocal_name, units=units,
                                                                quantity='reciprocal distance',
                                                                dimension_type=sidpy.DimensionType.RECIPROCAL))
                reciprocal_name = chr(ord(reciprocal_name) + 1)
            else:
                units = 'counts'
                dataset.set_dimension(int(dim), sidpy.Dimension(values, name=spatial_name, units=units,
                                                                quantity='distance',
                                                                dimension_type=sidpy.DimensionType.SPATIAL))
                spatial_name = chr(ord(spatial_name) + 1)

    # utility functions

    def __read_tag_group(self, tags):

        g_sorted = read_byte(self.__f)
        # is the group open?
        opened = read_byte(self.__f)
        # number of Tags
        if self.file_version == 3:
            n_tags = read_long(self.__f)
        else:
            n_tags = read_64_long(self.__f)

        # read Tags
        for i in range(n_tags):
            data = read_byte(self.__f)
            is_data = (data == 21)

            len_tag_label = read_short(self.__f)

            if len_tag_label > 0:
                tag_label = self.__f.read(len_tag_label).decode('latin-1')
            else:
                tag_label = '0'
                for key in tags:
                    if key.isdigit():
                        tag_label = str(int(key) + 1)
            if is_data:
                value = self.__read_any_data()
                tags[tag_label] = value
            else:
                tags[tag_label] = {}
                self.__read_tag_group(tags[tag_label])
        return 1

    def __encoded_type_size(self, et):
        # returns the size in bytes of the data type
        if et == 0:
            width = 0
        elif et in (BOOLEAN, CHAR, OCTET):
            width = 1
        elif et in (SHORT, USHORT):
            width = 2
        elif et in (LONG, ULONG, FLOAT):
            width = 4
        elif et == DOUBLE:
            width = 8
        else:
            # returns -1 for unrecognised types
            width = -1
        return width

    def __read_any_data(self):
        if self.file_version == 4:
            tag_byte_length = struct.unpack_from('>Q', self.__f.read(8))[0]
            # DM4 specifies this property as always big endian

        delim = read_string(self.__f, 4)
        if delim != b"%%%%":
            raise Exception(hex(self.__f.tell()) + ": Tag Type delimiter not %%%%")
        if self.file_version == 4:
            n_in_tag = read_64_long(self.__f)
        else:
            n_in_tag = read_long(self.__f)
        # higher level function dispatching to handling data types to other functions
        # - get Type category (short, long, array...)
        if self.file_version == 4:
            encoded_type = read_64_long(self.__f)
        else:
            encoded_type = read_long(self.__f)
        # - calc size of encoded_type
        et_size = self.__encoded_type_size(encoded_type)
        if et_size > 0:
            data = self.__read_native_data(encoded_type, et_size)
        elif encoded_type == STRING:
            string_size = read_long(self.__f)
            data = self.__read_string_data(string_size)
        elif encoded_type == STRUCT:
            struct_types = self.__read_struct_types()
            data = self.__read_struct_data(struct_types)
        elif encoded_type == ARRAY:
            array_types = self.__read_array_types()
            data = self.__read_array_data(array_types)
        else:
            raise Exception("rAnD, " + hex(self.__f.tell()) + ": Can't understand encoded type")
        return data

    def __read_native_data(self, encoded_type, et_size):
        # reads ordinary data types
        if encoded_type in readFunc.keys():
            val = readFunc[encoded_type](self.__f)
        else:
            raise Exception("rND, " + hex(self.__f.tell()) + ": Unknown data type " + str(encoded_type))
        return val

    def __read_string_data(self, string_size):
        # reads string data
        if string_size <= 0:
            r_string = ""
        else:
            # !!! *Unicode* string (UTF-16)... convert to Python unicode str
            r_string = read_string(self.__f, string_size)
            r_string = str(r_string, "utf_16_le")
        return r_string

    def __read_array_types(self):
        # determines the data types in an array data type
        array_type = read_long(self.__f)
        item_types = []
        if array_type == STRUCT:
            item_types = self.__read_struct_types()
        elif array_type == ARRAY:
            item_types = self.__read_array_types()
        else:
            item_types.append(array_type)
        return item_types

    def __read_array_data(self, array_types):
        # reads array data
        array_size = read_long(self.__f)
        item_size = 0
        encoded_type = 0
        for i in range(len(array_types)):
            encoded_type = int(array_types[i])
            et_size = self.__encoded_type_size(encoded_type)
            item_size += et_size
        buf_size = array_size * item_size

        if len(array_types) == 1 and encoded_type == USHORT \
                and array_size < 256:
            # treat as string
            val = self.__read_string_data(buf_size)
        else:
            # treat as binary data
            # - store data size and offset as tags
            val = self.__f.read(buf_size)
        return val

    def __read_struct_types(self):
        # analyses data types in a struct
        if self.file_version == 4:
            struct_name_length = read_64_long(self.__f)
            n_fields = read_64_long(self.__f)
        else:
            struct_name_length = read_long(self.__f)
            n_fields = read_long(self.__f)

        field_types = []
        name_length = 0
        for i in range(n_fields):
            if self.file_version == 4:
                name_length = read_64_long(self.__f)
                field_type = read_64_long(self.__f)
            else:
                name_length = read_long(self.__f)
                field_type = read_long(self.__f)
            field_types.append(field_type)
        return field_types

    def __read_struct_data(self, struct_types):
        # reads struct data based on type info in structType
        data = []
        for i in range(len(struct_types)):
            encoded_type = struct_types[i]
            et_size = self.__encoded_type_size(encoded_type)
            # get data
            data.append(self.__read_native_data(encoded_type, et_size))
        return data

    # ## END utility functions ###

    def get_filename(self):
        return self.__filename

    filename = property(get_filename)

    def get_tags(self):
        return self.__stored_tags

    tags = property(get_tags)

    def get_raw(self):
        """Extracts  data as np array"""

        # DataTypes for image data <--> PIL decoders
        data_types = {
            1: '<u2',  # 2 byte integer signed ("short")
            2: '<f4',  # 4 byte real (IEEE 754)
            3: '<c8',  # 8 byte complex (real, imaginary)
            4: '',  # ?
            # 4 byte packed complex (see below)
            5: (numpy.int16, {'real': (numpy.int8, 0), 'imaginary': (numpy.int8, 1)}),
            6: '<u1',  # 1 byte integer unsigned ("byte")
            7: '<i4',  # 4 byte integer signed ("long")
            # I do not have any dm3 file with this format to test it.
            8: '',  # rgb view, 4 bytes/pixel, unused, red, green, blue?
            9: '<i1',  # byte integer signed
            10: '<u2',  # 2 byte integer unsigned
            11: '<u4',  # 4 byte integer unsigned
            12: '<f8',  # 8 byte real
            13: '<c16',  # byte complex
            14: 'bool',  # 1 byte binary (ie 0 or 1)
            # Packed RGB. It must be a recent addition to the format because it does
            # not appear in http://www.microscopy.cen.dtu.dk/~cbb/info/dmformat/
            23: (numpy.float32,
                 {'R': ('<u1', 0), 'G': ('<u1', 1), 'B': ('<u1', 2), 'A': ('<u1', 3)}),
        }

        # find main image
        for key in self.__stored_tags['ImageList']:

            if key.isdigit():
                if 'ImageData' in self.__stored_tags['ImageList'][key]:
                    if 'Data' in self.__stored_tags['ImageList'][key]['ImageData'] \
                            and 'DataType' in self.__stored_tags['ImageList'][key]['ImageData'] \
                            and 'Dimensions' in self.__stored_tags['ImageList'][key]['ImageData']:
                        if int(key) > self.__chosen_image:
                            self.__chosen_image = int(key)
        if self.__chosen_image < 0:
            raise IOError('Did not find data in file')

        # get relevant Tags
        byte_data = self.__stored_tags['ImageList'][str(self.__chosen_image)]['ImageData']['Data']
        data_type = self.__stored_tags['ImageList'][str(self.__chosen_image)]['ImageData']['DataType']
        dimensions = self.__stored_tags['ImageList'][str(self.__chosen_image)]['ImageData']['Dimensions']

        # get shape from Dimensions
        shape = []
        for dim in dimensions:
            shape.append(dimensions[dim])

        # get data_type and reformat into numpy array
        dt = data_types[data_type]
        if dt == '':
            raise TypeError('The datatype is not supported')
        else:
            raw_data = numpy.frombuffer(byte_data, dtype=dt, count=numpy.cumprod(shape)[-1]).reshape(shape, order='F')
        # delete byte data in dictionary
        self.__stored_tags['ImageList'][str(self.__chosen_image)]['ImageData']['Data'] = 'read'
        return raw_data

    data_cube = property(get_raw)


if __name__ == '__main__':
    pass  # print "DM3lib v.%s"%version



# -*- coding: utf-8 -*-
"""
Created on Tue Oct 20 15:39:40 2020

@author: 4sv
"""

from __future__ import division, print_function, absolute_import, unicode_literals
import h5py
import warnings
import sidpy
from sidpy.sid import Reader
import numpy as np

try:
    import pyUSID as usid
except ModuleNotFoundError:
    usid = None


class Usid_reader(Reader):
    
    
    debugLevel = 0
    
    def __init__(self, file_path, verbose=False):
        
        """
        Parameters
        ----------
        
        file_path : str 
            path to the file containing USID Dataset.
        verbose: bool
        
        Attributes
        ----------
        self._input_file_path: str
            path provided to the file containing USID Dataset
        self._file: HDF5 file opened in 'r' mode
        
        
        Returns
        -------
        NONE (Initializes the class attributes)
        """
        
        super(Usid_reader, self).__init__(file_path)
        self.verbose = verbose
        
        if not usid:
            raise ModuleNotFoundError('Please install pyUSID to use this Reader')
        try:
            self._file = h5py.File(self._input_file_path, mode = 'r')
        except FileNotFoundError:
            raise FileNotFoundError('File not found')
        
        
        
        
    def _get_maindatasets(self):
        """
        Gets the maindatasets in the paths provided
        If the dataset path is not provided, gets all the datasets in the file 
        using usid.hdf_utils.get_all_main()
        
        Raises an error if no main datasets are found
        
        Attributes
        ----------
        self._main_datasets: list
            list of all main USID datasets found in the file path
        
        Returns
        -------
        NONE (writes all the maindatasets in the path to attribute self._main_datasets)
        """
        
        if self._dataset_path is not None:
            self._main_datasets = []
            
            if type(self._dataset_path) == list:
                for dataset in self._dataset_path:
                    if usid.hdf_utils.check_if_main(self._file[dataset]):
                        self._main_datasets.append(usid.USIDataset(self._file[dataset]))
                    else:
                        warnings.warn('{} is not a main dataset'.format(dataset))
                        
            elif type(self._dataset_path) == str:
                if usid.hdf_utils.check_if_main(self._file[self._dataset_path]):
                    self._main_datasets.append(usid.USIDataset(self._file[self._dataset_path]))
                else:
                    warnings.warn('{} is not a main dataset'.format(self._dataset_path))
                
            
            else:
                raise TypeError('Provide a path (StringType) to the main dataset or a List of paths')
            
        else:
            self._main_datasets = usid.hdf_utils.get_all_main(self._file)
        
        
        
        if len(self._main_datasets) == 0:
            raise TypeError('There are no main USID datasets in this file')
        
        
        
    @staticmethod
    def _get_dimension_descriptors(main_dataset, verbose):
        
        """
        Parameters
        ----------
        main_dataset: USID main dataset
            
        verbose: bool
        
        Gets the dimension descriptors for the maindataset provided as the
        argument
        
        
        Returns
        -------
        dims: list
        dim_labels: list
        dim_types: list
            spectral or spatial
        dim_quantities: list
        dim_units: list
            list of units of each dimension
        dim_values: list
            list of values at which the readings are recorded
        """
        
        pos_dim, spec_dim = 0, 0
        dims, dim_labels, dim_types, dim_quantities, dim_units, dim_values = [], [], [], [], [], []
        for dim, dim_label in enumerate(main_dataset.n_dim_labels):
            dims.append(dim)
            dim_labels.append(dim_label)
            if dim_label in main_dataset.pos_dim_labels:
                dim_types.append('spatial')
                dim_values.append(main_dataset.get_pos_values(dim_label))
                descriptor = main_dataset.pos_dim_descriptors[pos_dim].split('(')
                if descriptor[1][0] != ')':
                    dim_quantities.append(descriptor[0][:-1])
                    dim_units.append(descriptor[1][:-1])
                else:
                    dim_quantities.append(descriptor[0][:-1])
                    dim_units.append('generic')
                pos_dim += 1
                
            else:
                dim_types.append('spectral')
                dim_values.append(main_dataset.get_spec_values(dim_label))# error here
                descriptor = main_dataset.spec_dim_descriptors[spec_dim].split('(')
                if descriptor[1][0] != ')':
                    dim_quantities.append(descriptor[0][:-1])
                    dim_units.append(descriptor[1][:-1])
                else:
                    dim_quantities.append(descriptor[0][:-1])
                    dim_units.append('generic')
                spec_dim += 1
            if verbose:
                print('Read dimension {} of  type {} as {} ({})'.format(dim_labels[dim], dim_types[dim],
                                                                    dim_quantities[dim], dim_units[dim]))
        
        return dims, dim_labels, dim_types, dim_quantities, dim_units, dim_values
    
    @staticmethod
    def _get_main_data_descriptors(main_dataset):
        
        """
        Gets the descriptors (quatity and units) of the output quantity
        of a non compund main dataset
        Parameters
        ----------
        main_dataset: USID main dataset
        Returns
        -------
        qunatity: str
            Name of the output quantity
        Units: str
        """
        
        des = main_dataset.data_descriptor.split('(')
        try:
            des[1]
            quantity = des[0][:-1]
            units = des[1][:-1]
        except IndexError:
            quantity = des[0]
            units = 'generic'
            
        return quantity, units
    
    @staticmethod
    def _get_compound_data_descriptors(name):
        """
        Gets the descriptors (quatity and units) of the output quantity
        of a compund dataset
        
        Parameters
        ----------
        name: str
        
        Returns
        -------
        qunatity: str 
            Name of the output quantity 
        Units: str
        """
        
        des = name.split('[')
        try:
            des[1]
            quantity = des[0][:-1]
            units = des[1][:-1]
        except IndexError:
            quantity = des[0]
            units = 'generic'
            
        return quantity, units
    
    def _get_metadata(main_dataset):
        metadata = {}
        #We get all the attributes corresponding to main_dataset
        atts = sidpy.hdf_utils.get_attributes(main_dataset)
        for attr_name in (atts):
            #We omit h5py reference files
            if not isinstance(atts[attr_name], (h5py.h5r.Reference, h5py.h5r.RegionReference)):
                #We do not want quantity of units either
                if not (attr_name != 'quantity' or attr_name != 'units'):
                    metadata[attr_name] = atts[attr_name]
                    
        return metadata
        
        
    
    def read(self, dataset_path = None):
        """
        Parameters
        ----------
        dataset_path: str or list
            path inside the file to the main Dataset or a list of paths
        
        Returns
        -------
        sid_datasets: sidpy dataset or list
        sidpy dataset object when a single dataset is found in the path or list 
        of sidpy dataset objects
        """
        self._dataset_path = dataset_path
        self._get_maindatasets()
        
        def read_not_in_ndim():
            h5_file = self._file
            h5_main = usid.hdf_utils.find_dataset(h5_file, 'Raw_Data')[0]
            data_array = np.array(h5_main)# (3600, 216576)
            position_index = np.array(h5_main.h5_pos_inds)# (3600, 2)
            position_values = np.array(h5_main.h5_pos_vals)# (3600, 2)
            spectroscopic_index = np.array(h5_main.h5_spec_inds)# (216576, 2)
            spectroscopic_values = np.array(h5_main.h5_spec_vals)# (216576, 2)

            metadata_dict = {"pos_index": position_index, "pos_values": position_values, 
                            "spec_index": spectroscopic_index, "spec_values": spectroscopic_values}
            sid_data = sidpy.Dataset.from_array(data_array)
            sid_data.metadata = metadata_dict
            return sid_data

        sid_datasets = []
        for j,main_dataset in enumerate(self._main_datasets):
            # Check if the main dataset is compound
            if main_dataset.dtype.names is not None:
                #Get descriptors of the main dataset
                dims, dim_labels, dim_types, dim_quantities, dim_units, dim_values = Usid_reader._get_dimension_descriptors(main_dataset, verbose = self.verbose)
                
                for name in (main_dataset.dtype.names):
                    sid_dataset = sidpy.Dataset.from_array(main_dataset.get_n_dim_form()[name])
                    #Get descriptors (units) of the output quantity
                    sid_dataset.quantity, sid_dataset.units = Usid_reader._get_compound_data_descriptors(name)
                    
                    #Set dimensions for the main dataset
                    for i in range(len(dim_labels)):
                        sid_dataset.set_dimension(dims[i], sidpy.Dimension(dim_values[i],name = dim_labels[i],
                                                       units=dim_units[i],
                                                       quantity=dim_quantities[i],
                                                       dimension_type=dim_types[i]))
                    
                    # Dealing with metadata
                    meta_data = Usid_reader._get_metadata(main_dataset)
                    sid_dataset.metadata.update(meta_data)
                    sid_datasets.append(sid_dataset)
            
            #For a non-compound dataset
            else:
                #Get descriptors of the main dataset
                try :
                    dims, dim_labels, dim_types, dim_quantities, dim_units, dim_values = Usid_reader._get_dimension_descriptors(main_dataset, verbose = self.verbose)
                except ValueError:
                    return read_not_in_ndim()
                    
                sid_dataset = sidpy.Dataset.from_array(main_dataset.get_n_dim_form())
                #Get descriptors (units) of each of the output quantity
                sid_dataset.quantity, sid_dataset.units = Usid_reader._get_main_data_descriptors(main_dataset)

                
                #Set dimensions for the main dataset
                for i in range(len(dim_labels)):
                    sid_dataset.set_dimension(dims[i], sidpy.Dimension(dim_values[i],name = dim_labels[i],
                                                       units=dim_units[i],
                                                       quantity=dim_quantities[i],
                                                       dimension_type=dim_types[i]))
                
                meta_data = Usid_reader._get_metadata(main_dataset)
                sid_datasets.append(sid_dataset)
                    
            
                

        
        if len(sid_datasets) == 1:
            sid_datasets = sid_datasets[0]
            

        return sid_datasets

    def close(self):
        """
        Closes the HDF5 file.
        """
        if self._file is not None:
            self._file.close()
            self._file = None


"""
Created on Fri Jan 25 2024

@author: Boris Slautin
"""

import numpy as np  # For array operations
import base64
import struct
import sidpy as sid
from sidpy.sid import Reader

#XML reader
import xml.etree.ElementTree as ET

import gzip


class AxzReader(Reader):
    """
    Extracts data and metadata from Analysis Studio AFM-IR files (.axz) files containing
    images or spectra
    """
    def __init__(self, file_path, *args, **kwargs):
        # if apt == None:
        #     raise ModuleNotFoundError('You attempted to load an .axz file, but this requires anasyspythontools.\n \
        #     Please Load it with pip install anasyspythontools , restart and retry')

        super().__init__(file_path, *args, **kwargs)

        self.datasets = {}

    def read(self, verbose=False):
        '''
        Reads the file given in file_path into a sidpy dataset

        Parameters
        ----------
        verbose : Boolean (Optional)
            Whether or not to show  print statements for debugging

        Returns
        -------
        sidpy.Dataset : Dict with sidpy.Dataset objects.
                    Multi-channel inputs are separated into individual dataset objects
        '''

        file_path = self._input_file_path

        _file = self._file_to_dict(file_path)

        _images  = _file['HeightMaps']
        self._images = _images
        _spectra = _file['RenderedSpectra']

        self._file = _file

        channel_number = 0
        if (_images is not None) and (len(_images) > 0):
            for key in _images.keys():
                dataset = self._read_image(_images[key])
                key_channel = f"Channel_{int(channel_number):03d}"
                self.datasets[key_channel] = dataset
                channel_number += 1

        if (_spectra is not None) and (len(_spectra) > 0):
            for key in _spectra.keys():
                dataset = self._read_spectrum(_spectra[key])
                key_channel = f"Channel_{int(channel_number):03d}"
                self.datasets[key_channel] = dataset
                channel_number += 1

        return self.datasets

    def _file_to_dict(self, filepath):
        with gzip.open(filepath) as f:
            f_data = ET.iterparse(f)
            for event, el in f_data:
                el.tag = el.tag.split('}', -1)[-1]

        data = f_data.root

        res_dict = {'HeightMaps':{},
                    'RenderedSpectra':{}}

        for items in data:
            if items.tag == 'HeightMaps':
                for i, item in enumerate(items):
                    res_dict['HeightMaps'][i] = self._xml_to_dict(item)
            elif items.tag == 'RenderedSpectra':
                for i, item in enumerate(items):
                    res_dict['RenderedSpectra'][i] = self._xml_to_dict(item)

        return res_dict#self._xml_to_dict(data)

    def _xml_to_dict(self, data):
        full_res = {}
        for attr in data.items():
            ET.SubElement(data, attr[0])
            data.find(attr[0]).text = attr[1]
        if len(list(data)) == 0:
            if '64' in data.tag:
                return self.decode_bs64(data.text)
            else:
                return data.text
        for items in data:
            attr = self._return_layer_attr(items)  # return_child_dict(items)
            if attr:
                full_res.update(attr)
            if items.tag is not None:
                el1 = self._xml_to_dict(items)
                full_res[items.tag] = el1

        return full_res

    @staticmethod
    def decode_bs64(data):
        """Returns base64 data decoded in a numpy array"""
        if data is None:
            return np.ndarray(0)
        decoded_bytes = base64.b64decode(data.encode())
        fmt = 'f' * int((len(decoded_bytes) / 4))
        structured_data = struct.unpack(fmt, decoded_bytes)
        decoded_array = np.array(structured_data)
        return decoded_array

    @staticmethod
    def _return_layer_attr(el_tree):
        if len(el_tree.items()) > 0:
            res = {}
            for elem in el_tree.items():
                res[elem[0]] = elem[1]
            return res
        else:
            return None

    def _read_image(self, obj):
        # Convert it to sidpy dataset object
        title = obj['Label']
        units = str(obj['UnitPrefix']) + str(obj['Units'])

        sizeX, sizeY = float(obj['Size']['X']), float(obj['Size']['Y'])
        resX, resY = int(obj['Resolution']['X']), int(obj['Resolution']['Y'])

        data = obj['SampleBase64'].reshape(resX,resY)

        data_set = sid.Dataset.from_array(data)
        data_set.data_type = 'image'
        data_set.title = title
        data_set.units = units

        data_set.set_dimension(0, sid.Dimension(np.linspace(0, sizeX , resX), 'x'))
        data_set.x.dimension_type = 'spatial'
        data_set.x.quantity = 'distance'
        data_set.x.units = 'um'

        data_set.set_dimension(1, sid.Dimension(np.linspace(0, sizeY , resY), 'y'))
        data_set.y.dimension_type = 'spatial'
        data_set.y.quantity = 'distance'
        data_set.y.units = 'um'

        data_set.metadata = self._image_metadata(obj)

        return data_set

    def _image_metadata(self, obj):
        metadata = {}

        for key in obj['Tags']:
            metadata[key] = obj['Tags'][key]

        excluded_keys = ['Tags', 'Visible','SampleBase64']

        for key in obj:
            if key not in excluded_keys:
                metadata[key] = obj[key]

        return metadata

    def _read_spectrum(self, obj):
        data   = obj['DataChannels']['SampleBase64']

        _start_wn = float(obj['StartWavenumber'])
        _end_wn = float(obj['EndWavenumber'])
        _datapoints = int(obj['DataPoints'])
        data_x = np.linspace(_start_wn, _end_wn, _datapoints)
        title = obj['Label']

        posX, posY = float(obj['Location']['X']), float(obj['Location']['Y'])

        data_set = sid.Dataset.from_array(data)
        data_set.title = title
        data_set.data_type = 'spectrum'
        data_set.quantity = 'intensity'
        data_set.units = 'a.u.'

        data_set.set_dimension(0, sid.Dimension(data_x, 'wavenumber'))
        data_set.dim_0.dimension_type = 'spectral'
        data_set.dim_0.quantity = 'wavenumber'
        data_set.dim_0.units = 'cm-1'#?

        data_set.metadata = self._spectrum_metadata(obj)
        return data_set

    def _spectrum_metadata(self, obj):
        metadata = {}
        excluded_keys = ['Visible',
                         '{http://www.w3.org/2001/XMLSchema-instance}nil',
                         'DataChannels',
                         'Polarization',
                         'DutyCycle',
                         'PulseRate']

        for key in obj:
            if key not in excluded_keys:
                metadata[key] = obj[key]

        return metadata


"""
Tools to read, data in from generic files

Submodules
----------

.. autosummary::
    :toctree: _autosummary

    ImageReader
"""

from .image import ImageReader

__all__ = ['ImageReader']

all_readers = [ImageReader]



"""
Part of SciFiReaders, a pycroscopy module
Author: Gerd Duscher
################################################################################
# Python class for reading GATAN DM3/DM4 (DigitalMicrograph) files
# and extracting all metadata
# --
# tested on EELS spectra, spectrum images and single-image files and image-stacks
# --
# based on the DM3_Reader plug-in (v 1.3.4) for ImageJ by Greg Jefferis <jefferis@stanford.edu>
# http://rsb.info.nih.gov/ij/plugins/DM3_Reader.html
# --
# Python adaptation: Pierre-Ivan Raynal <raynal@med.univ-tours.fr>
# http://microscopies.med.univ-tours.fr/
#
# Extended for EELS support by Gerd Duscher, UTK 2012
# Rewritten for integration of sidpy 2020
# Rewritten for added support of DM4 2021
#    gleaned some ideas from https://github.com/jamesra/dm4reader
# Works for python 3
#
################################################################################
"""

from __future__ import division, print_function, absolute_import, unicode_literals
import os
import struct
import time
import warnings

import numpy as np

import sidpy

__version__ = '0.1beta'

DM4DataTypeDict = {2: {'num_bytes': 2, 'signed': True, 'type_format': 'h'},
                   3: {'num_bytes': 4, 'signed': True, 'type_format': 'i'},
                   4: {'num_bytes': 2, 'signed': False, 'type_format': 'H'},
                   5: {'num_bytes': 4, 'signed': False, 'type_format': 'I'},
                   6: {'num_bytes': 4, 'signed': True, 'type_format': 'f'},
                   # 8byte float
                   7: {'num_bytes': 8, 'signed': True, 'type_format': 'd'},
                   8: {'num_bytes': 1, 'signed': False, 'type_format': 'B'},
                   9: {'num_bytes': 1, 'signed': False, 'type_format': 'c'},
                   10: {'num_bytes': 1, 'signed': True, 'type_format': 'b'},
                   # 8 bit long-long for DM4
                   11: {'num_bytes': 8, 'signed': True, 'type_format': 'q'},
                   12: {'num_bytes': 8, 'signed': False, 'type_format': 'Q'}
                   }


# Utility functions
def read_header_dm(dm_file):
    """Read the header of a DM3/DM4 file."""
    dm_file.seek(0)
    # int.from_bytes(dm_file.read(4), byteorder='big')
    dm_version = struct.unpack_from('>I', dm_file.read(4))[0]
    if dm_version == 3:
        file_size = struct.unpack_from('>I', dm_file.read(8))[0]
        dm_header_size = 4 + 4 + 4
    elif dm_version == 4:
        file_size = struct.unpack_from('>Q', dm_file.read(8))[0]
        dm_header_size = 4 + 8 + 4
    else:
        raise TypeError('This is not a DM3 or DM4 File')
    byteorder = struct.unpack_from('>I', dm_file.read(4))[0]
    if byteorder == 1:
        endian = '>'  # little nedian
    else:
        endian = '<'  # big endian

    if dm_version == 4:  # not sure why
        if endian == '<':
            endian = '>'
        else:
            endian = '<'
    return dm_version, file_size, endian, dm_header_size


def _read_tag_name(dm_file):
    """Read the name of a tag from the DM file."""
    # DM specifies this property as always big endian
    tag_name_len = struct.unpack_from('>H', dm_file.read(2))[0]

    tag_name = '0'
    if tag_name_len > 0:
        data = dm_file.read(tag_name_len)
        try:
            tag_name = data.decode('utf-8', errors='ignore')
        except UnicodeDecodeError:
            tag_name = None

    return tag_name


def _read_tag_garbage_str(dm_file):
    """
    DM4 has four bytes of % symbols in the tag.  Ensure it is there.
    """
    garbage_str = dm_file.read(4).decode('utf-8')
    assert garbage_str == '%%%%'


def _read_tag_data_info(dm_file, dm_version):
    if dm_version == 3:
        tag_array_length = struct.unpack('>l', dm_file.read(4))[0]
        format_str = '>' + tag_array_length * 'l'  # Big endian signed long
        tag_array_types = struct.unpack_from(format_str, dm_file.read(4 * tag_array_length))
    else:
        tag_array_length = struct.unpack_from('>Q', dm_file.read(8))[0]
        # DM4 specifies this property as always big endian
        format_str = '>' + tag_array_length * 'q'  # Big endian signed long

        tag_array_types = struct.unpack_from(format_str, dm_file.read(8 * tag_array_length))

    return tag_array_types


def read_string(dm_file, length=1):
    """Read len bytes as a string in file f"""
    read_bytes = dm_file.read(length)
    str_fmt = '>' + str(length) + 's'
    return struct.unpack(str_fmt, read_bytes)[0]


class DMReader(sidpy.Reader):
    """
    Reader of Digital Micrograph image and spectroscopy data

    This reader reads (attribute read) all the different data in the file and
    returns it as a dictionary 
    of sidpy.Datasets

    Parameter:
    ---------
    file_path: str
        filepath to dm3 or dm4 file.

    Return:
    ------
    datasets: dict
        dictionary of sidpy datasets
    """
    def __init__(self, file_path, verbose=False):
        super().__init__(file_path)

        # initialize variables ##
        self.verbose = verbose
        self.__filename = file_path
        self.datasets = {}
        self.spectral_dim = False

        # - open file for reading
        try:
            self.__dm_file = open(self.__filename, 'rb')
        except FileNotFoundError:
            raise FileNotFoundError('File not found')

        # - create Tags repositories
        self.__stored_tags = {'DM': {}}

        self.dm_version, self.file_size, self.endian, self.dm_header_size = read_header_dm(self.__dm_file)

        if self.verbose:
            print("Header info.:")
            print("- file version:", self.dm_version)
            print("- little endian:", self.endian)
            print("- file size:", self.file_size, "bytes")

        # don't read but close file
        self.close()

    def close(self):
        """Close the DM file."""
        self.__dm_file.close()

    def read(self):
        """Read the DM file and extract metadata."""
        try:
            self.__dm_file = open(self.__filename, 'rb')
        except FileNotFoundError:
            raise FileNotFoundError('File not found')

        t1 = time.time()
        self.__dm_file.seek(self.dm_header_size)

        self.__stored_tags = {'DM': {'dm_version': self.dm_version, 'file_size': self.file_size},
                              'original_filename': self.filename}

        self.__read_tag_group(self.__stored_tags)

        if self.verbose:
            print(f"-- {len(self.__stored_tags)} Tags read --")

        if self.verbose:
            t2 = time.time()
            print(f"| parse DM3 file: {t2 - t1:.3g} s")
        start = 0
        if '1' in self.__stored_tags['ImageList']:
            start=1
        channel_number = 0
        for image_number in self.__stored_tags['ImageList'].keys():
            if int(image_number) >= start:
                dataset = self.get_dataset(self.__stored_tags['ImageList'][image_number])
                if isinstance(dataset, sidpy.Dataset):
                    dataset.original_metadata['DM'] = self.__stored_tags['DM']
                    # convert linescan to spectral image
                    if self.spectral_dim and dataset.ndim == 2:
                        old_dataset = dataset.copy()
                        meta = dataset.original_metadata.copy()
                        basename = dataset.name
                        data = np.array(dataset).reshape(dataset.shape[0], 1, dataset.shape[1])
                        dataset = sidpy.Dataset.from_array(data, name=basename)
                        dataset.original_metadata = meta
                        dataset.set_dimension(0, old_dataset.dim_0)

                        dataset.set_dimension(1, sidpy.Dimension([1], name='y', units='pixels',
                                                                 quantity='distance',
                                                                 dimension_type='spatial'))
                        dataset.set_dimension(2, old_dataset.dim_1)
                        dataset.data_type = sidpy.DataType.SPECTRAL_IMAGE  # 'linescan'

                    dataset.quantity = 'intensity'
                    dataset.units = 'counts'
                    # dataset.title = basename
                    dataset.modality = 'generic'
                    dataset.source = 'SciFiReaders.DMReader'
                    dataset.original_metadata['DM']['full_file_name'] = self.__filename

                    key = f"Channel_{int(channel_number):03d}"
                    channel_number += 1
                    self.datasets[key] = dataset
                    self.extract_crucial_metadata(key)

        del self.__stored_tags['ImageList']
        main_dataset_key = list(self.datasets.keys())[0]

        for key, dataset in self.datasets.items():
            if 'urvey' in dataset.title:
                main_dataset_key = key
        if self.verbose:
            print(main_dataset_key)
        self.datasets[main_dataset_key].original_metadata.update(self.__stored_tags)
        self.close()
        return self.datasets

    def get_dataset(self, image_dict)->sidpy.Dataset:
        """
        Reads dictionary of imageList into sidpy.Dataset 
        """
        dataset = None
        _, file_name = os.path.split(self.__filename)
        basename, _ = os.path.splitext(file_name)
        if image_dict.get('ImageData', {}).get('Data', {}):
            dataset = sidpy.Dataset.from_array(self.get_raw(image_dict), name=basename)
            dataset.title = image_dict['Name']
            dataset.original_metadata = image_dict.copy()
            self.set_dimensions(dataset)
            self.set_data_type(dataset)
        return dataset

    def set_data_type(self, dataset):
        """Sets the data type of the dataset based on its shape and metadata."""
        self.spectral_dim = spectral_dim = len(dataset.get_spectral_dims()) > 0

        dataset.data_type = 'unknown'
        if dataset.original_metadata.get('ImageTags', {}).get('SI', {}):
            if len(dataset.shape) == 3:
                dataset.data_type = sidpy.DataType.SPECTRAL_IMAGE
            else:
                if spectral_dim:
                    dataset.data_type = sidpy.DataType.SPECTRAL_IMAGE  # 'linescan'
                else:
                    dataset.data_type = sidpy.DataType.IMAGE
                    dataset.metadata['image_type'] = 'survey image'

        if dataset.data_type == sidpy.DataType.UNKNOWN:
            if len(dataset.shape) > 3:
                raise NotImplementedError('Data_type not implemented yet')
            if len(dataset.shape) == 3:
                if spectral_dim:
                    dataset.data_type = sidpy.DataType.SPECTRAL_IMAGE
                else:
                    dataset.data_type = 'image_stack'
            elif len(dataset.shape) == 2:
                if spectral_dim:
                    # basename = dataset.name
                    dataset.data_type = sidpy.DataType.SPECTRAL_IMAGE
                else:
                    dataset.data_type = 'image'
            elif len(dataset.shape) == 1:
                if spectral_dim:
                    dataset.data_type = sidpy.DataType.SPECTRUM
                else:
                    dataset.data_type = sidpy.DataType.LINE_PLOT

    def set_dimensions(self, dataset):
        """Sets the dimensions of the dataset."""
        dimensions_dict = dataset.original_metadata.get('ImageData', {}).get('Calibrations', {})
        dimensions_dict = dimensions_dict.get('Dimension', {})
        reciprocal_name = 'u'
        spatial_name = 'x'

        for dim, dimension_tags in dimensions_dict.items():
            # Fix annoying scale of spectrum_images in Zeiss  and SEM images
            if dimension_tags['Units'] == '�m':
                dimension_tags['Units'] = 'nm'
                dimension_tags['Scale'] *= 1000.0

            if dimension_tags['Units'].strip() == '':
                units = 'counts'
            else:
                units = dimension_tags['Units']

            values = np.arange(dataset.shape[int(dim)]) - dimension_tags['Origin']
            values *= dimension_tags['Scale']

            if 'eV' == units:
                dataset.set_dimension(int(dim),
                                      sidpy.Dimension(values, name='energy_loss',
                                                      units=units, quantity='energy-loss',
                                                      dimension_type=sidpy.DimensionType.SPECTRAL))
            elif 'eV' in units:
                dataset.set_dimension(int(dim),
                                      sidpy.Dimension(values, name='energy',
                                                      units=units, quantity='energy',
                                                      dimension_type=sidpy.DimensionType.SPECTRAL))
            elif '1/' in units or units in ['mrad', 'rad']:
                dataset.set_dimension(int(dim),
                                      sidpy.Dimension(values, name=reciprocal_name,
                                                      units=units, quantity='reciprocal distance',
                                                      dimension_type=sidpy.DimensionType.RECIPROCAL))
                reciprocal_name = chr(ord(reciprocal_name) + 1)
            elif 'm' in units:
                dataset.set_dimension(int(dim),
                                      sidpy.Dimension(values, name=spatial_name, units=units,
                                                      quantity='distance',
                                                      dimension_type=sidpy.DimensionType.SPATIAL))
                spatial_name = chr(ord(spatial_name) + 1)
            else:
                units = 'frame'
                dataset.set_dimension(int(dim),
                                      sidpy.Dimension(values, name=spatial_name,
                                                      units=units, quantity='number',
                                                      dimension_type=sidpy.DimensionType.TEMPORAL))
                spatial_name = chr(ord(spatial_name) + 1)

        # For ill defined DM data
        if dataset.data_type.name == 'IMAGE':
            dataset.x.dimension_type = 'SPATIAL'
            dataset.y.dimension_type = 'SPATIAL'

    # utility functions

    def __read_tag_group(self, tags):
        """Reads a group of tags from the DM file."""
        _ = struct.unpack_from(self.endian + 'b', self.__dm_file.read(1))[0]
        _ = struct.unpack_from(self.endian + 'b', self.__dm_file.read(1))[0]

        if self.dm_version == 3:
            # this property is always big endian
            num_tags = struct.unpack_from('>l', self.__dm_file.read(4))[0]
        else:
             # this property is always big endian
            num_tags = struct.unpack_from('>Q', self.__dm_file.read(8))[0]

        # read Tags
        for _ in range(num_tags):
            tag_type = struct.unpack_from(self.endian + 'B', self.__dm_file.read(1))[0]
            is_data = tag_type == 21

            tag_label = _read_tag_name(self.__dm_file)
            if self.dm_version == 4:
                _ = struct.unpack_from('>Q', self.__dm_file.read(8))[0]
            if tag_label == '0':
                for key in tags:
                    if key.isdigit():
                        tag_label = str(int(key) + 1)

            if is_data:
                value = self.__read_any_data()
                tags[tag_label] = value
            else:
                tags[tag_label] = {}
                self.__read_tag_group(tags[tag_label])
        return 1

    def __read_any_data(self):
        """Reads any data from the DM file."""
        _read_tag_garbage_str(self.__dm_file)
        tag_array_types = _read_tag_data_info(self.__dm_file, self.dm_version)
        encoded_type = tag_array_types[0]

        if encoded_type < 13:
            data = self.__read_native_data(encoded_type)
        elif encoded_type == 18:  # STRING
            data = self.__read_string_data(tag_array_types[1])
        elif encoded_type == 15:  # STRUCT:
            data = self.__read_struct_data(tag_array_types)
        elif encoded_type == 20:  # ARRAY:
            if tag_array_types[1] == 15:
                data = []
                for _ in range(tag_array_types[-1]):
                    data.append(self.__read_struct_data(tag_array_types[1:]))
            elif tag_array_types[1] == 20:
                data = self.__read_array_data(tag_array_types[1:])
            else:
                data = self.__read_array_data(tag_array_types)
        else:
            out_error = "rAnD, " + str(self.__dm_file.tell()) + ": Can't understand encoded type"
            raise ValueError(out_error)
        return data

    def __read_native_data(self, encoded_type):
        """Reads native data from the DM file."""
        if encoded_type in DM4DataTypeDict:
            data_type = DM4DataTypeDict[encoded_type]
            format_str = self.endian + data_type['type_format']
            byte_data = self.__dm_file.read(data_type['num_bytes'])
            val = struct.unpack_from(format_str, byte_data)[0]
        else:
            out_error = "rND, " + hex(self.__dm_file.tell()) + ": Unknown data type " + str(encoded_type)
            raise ValueError(out_error)
        return val

    def __read_string_data(self, string_size):
        """Reads string data from the DM file."""
        if string_size <= 0:
            r_string = ""
        else:
            # !!! *Unicode* string (UTF-16)... convert to Python unicode str
            r_string = read_string(self.__dm_file, string_size)
            r_string = str(r_string, "utf_16_le")
        return r_string

    def __read_array_data(self, array_types):
        """Reads array data from the DM file."""
        array_size = array_types[-1]
        item_size = 0
        encoded_type = 0
        for i in range(len(array_types)-2):
            encoded_type = int(array_types[i+1])
            et_size = DM4DataTypeDict[encoded_type]['num_bytes']
            item_size += et_size
        buf_size = array_size * item_size

        if len(array_types)-2 == 1 and encoded_type == 4 and array_size < 256:
            # treat as string
            val = self.__read_string_data(buf_size)
        else:
            # treat as binary data
            # - store data size and offset as tags
            val = self.__dm_file.read(buf_size)
        return val

    def __read_struct_data(self, struct_types):
        """Reads struct data from the DM file."""
        data = []
        for encoded_type in struct_types[4::2]:
            data.append(self.__read_native_data(encoded_type))
        return data

    # ## END utility functions ###

    def get_filename(self):
        """Gets the filename of the DM file."""
        return self.__filename

    filename = property(get_filename)

    def get_tags(self):
        """Gets the tags from the DM file."""
        return self.__stored_tags

    tags = property(get_tags)

    def get_raw(self, image_dict):
        """Extracts  data as np array"""

        # DataTypes for image data <--> PIL decoders
        data_types = {
            1: '<u2',  # 2 byte integer signed ("short")
            2: '<f4',  # 4 byte real (IEEE 754)
            3: '<c8',  # 8 byte complex (real, imaginary)
            4: '',  # ?
            # 4 byte packed complex (see below)
            5: (np.int16, {'real': (np.int8, 0), 'imaginary': (np.int8, 1)}),
            6: '<u1',  # 1 byte integer unsigned ("byte")
            7: '<i4',  # 4 byte integer signed ("long")
            # I do not have any dm3 file with this format to test it.
            8: '',  # rgb view, 4 bytes/pixel, unused, red, green, blue?
            9: '<i1',  # byte integer signed
            10: '<u2',  # 2 byte integer unsigned
            11: '<u4',  # 4 byte integer unsigned
            12: '<f8',  # 8 byte real
            13: '<c16',  # byte complex
            14: 'bool',  # 1 byte binary (ie 0 or 1)
            # Packed RGB. It must be a recent addition to the format because it does
            # not appear in http://www.microscopy.cen.dtu.dk/~cbb/info/dmformat/
            23: (np.float32, {'R': ('<u1', 0), 'G': ('<u1', 1), 'B': ('<u1', 2), 'A': ('<u1', 3)}),
        }

        # get relevant Tags
        byte_data = image_dict['ImageData']['Data']
        data_type = image_dict['ImageData']['DataType']
        dimensions = image_dict['ImageData']['Dimensions']

        # get shape from Dimensions
        shape = []
        for dim in dimensions:
            shape.append(dimensions[dim])

        # get data_type and reformat into numpy array
        dt = data_types[data_type]
        if dt == '':
            raise TypeError('The datatype is not supported')
        raw_data = np.frombuffer(byte_data, dtype=dt, 
                                 count=np.cumprod(shape)[-1]).reshape(shape, order='F')
        # delete byte data in dictionary
        image_dict['ImageData']['Data'] = 'read'
        return raw_data

    def extract_crucial_metadata(self, dataset_key):
        """Extracts crucial metadata from the original_metadata dictionary."""
        original_metadata = self.datasets[dataset_key].original_metadata
        if not isinstance(original_metadata, dict):
            raise TypeError('We need a dictionary to read')
        experiment = {}
        if not original_metadata.get('DM', {}):
            return experiment
        if 'ImageTags' not in original_metadata:
            return experiment
        exp_dictionary = original_metadata.get('ImageTags', {})

        tag = exp_dictionary.get('EELS', {}).get('Acquisition', {}).get('Exposure (s)', 0.)
        experiment['single_exposure_time'] = tag
        tag = exp_dictionary.get('EELS', {}).get('Integration', {}).get('Integration (s)', -1)
        experiment['exposure_time'] = tag
        for key, item in exp_dictionary.get('EELS', {}).get('Acquisition', {}).items():
            if 'frames' in key:
                experiment['number_of_frames'] = item
                print('frames: ',key)
        exp_dictionary = original_metadata.get('ImageTags', {}).get('EELS', {})
        exp_dictionary = exp_dictionary.get('Experimental Conditions', {})
        experiment['convergence_angle'] = exp_dictionary.get('Convergence Angle', 0.)
        experiment['collection_angle'] = exp_dictionary.get('Collection Angle', 0.)
        experiment['number_of_frames'] = experiment.get('number_of_frames', 1)
        if 'exposure_time' not in experiment:
            experiment['exposure_time'] = experiment.get('number_of_frames', 1) \
                                        * experiment.get('single_exposure_time', -1)
        if experiment['exposure_time'] < 0:
            exp_dictionary = original_metadata.get('ImageTags', {}).get('Acquisition', {})
            exp_dictionary = exp_dictionary.get('Parameters', {}).get('High Level', {})
            experiment['exposure_time'] = exp_dictionary.get('Exposure (s)', 0)

        exp_dictionary = original_metadata.get('ImageTags', {}).get('Microscope Info', {})
        experiment['microscope'] = exp_dictionary.get('Microscope', 'Unknown')
        experiment['acceleration_voltage'] = exp_dictionary.get('Voltage', None)

        self.datasets[dataset_key].metadata['experiment']  = experiment
        return experiment

class DM3Reader(DMReader):
    """
    DM3 file reader class.
    """
    def __init__(self, file_path, verbose=False):

        warnings.warn(DeprecationWarning('Use DMReader class instead marking\n '
                                         'Note that you can now read dm4 files too'))
        super().__init__(file_path, verbose=verbose)



from .igor_ibw import IgorIBWReader, IgorMatrixReader
from .AR_hdf5  import ARhdf5Reader
from .bruker_nano import BrukerAFMReader
from .gwyddion import GwyddionReader
from .axz import AxzReader
from .mdt import MDTReader
from .wsxm import WSxM1DReader, WSxM2DReader, WSxM3DReader

__all__ = ['IgorIBWReader', 'ARhdf5Reader', 'BrukerAFMReader', 'GwyddionReader', 'IgorMatrixReader', 'AxzReader',
           'WSxM1DReader', 'WSxM2DReader', 'WSxM3DReader','MDTReader']
all_readers = [IgorIBWReader, ARhdf5Reader, BrukerAFMReader, GwyddionReader, IgorMatrixReader, AxzReader,
               WSxM1DReader, WSxM2DReader, WSxM3DReader,MDTReader]



"""
Hyperspy dataset converter to sidpy
part of SciFiReader, a pycroscopy package

author: Gerd Duscher, UTK
First Version 11/19/2021
"""

import sidpy
import numpy as np

try:
    import hyperspy.api as hs
except ModuleNotFoundError:
    hs = None


def convert_hyperspy(s):
    """
    imports a hyperspy signal object into sidpy.Dataset

    Parameters
    ----------
    s: hyperspy dataset

    Return
    ------
    dataset: sidpy.Dataset
    """
    if not hs:
        raise ModuleNotFoundError("Hyperspy is not installed")
    if not isinstance(s, (hs.signals.Signal1D, hs.signals.Signal2D)):
        raise TypeError('This is not a hyperspy signal object')
    dataset = sidpy.Dataset.from_array(s, name=s.metadata.General.title)
    # Add dimension info
    axes = s.axes_manager.as_dictionary()

    if isinstance(s, hs.signals.Signal1D):
        if s.data.ndim < 2:
            dataset.data_type = 'spectrum'
        elif s.data.ndim > 1:
            if s.data.ndim == 2:
                dataset = sidpy.Dataset.from_array(np.expand_dims(s, 2), name=s.metadata.General.title)
                dataset.set_dimension(2, sidpy.Dimension([0], name='y', units='pixel',
                                      quantity='distance', dimension_type='spatial'))
            dataset.data_type = sidpy.DataType.SPECTRAL_IMAGE
        for key, axis in axes.items():
            if axis['navigate']:
                dimension_type = 'spatial'
            else:
                dimension_type = 'spectral'
            if 'scale' not in axis:
                axis['scale'] = 1.0
            if 'offset' not in axis:
                axis['offset'] = 0.0    
            dim_array = np.arange(axis['size']) * axis['scale'] + axis['offset']
            if axis['units'] == '':
                axis['units'] = 'frame'
            dataset.set_dimension(int(key[-1]), sidpy.Dimension(dim_array, name=axis['name'], units=axis['units'],
                                                                quantity=axis['name'], dimension_type=dimension_type))

    elif isinstance(s, hs.signals.Signal2D):
        if s.data.ndim < 4:
            if s.data.ndim == 2:
                dataset.data_type = 'image'
            elif s.data.ndim == 3:
                dataset.data_type = 'image_stack'

            for key, axis in axes.items():
                if axis['navigate']:
                    dimension_type = 'temporal'
                else:
                    dimension_type = 'spatial'
                dim_array = np.arange(axis['size']) * axis['scale'] + axis['offset']
                if axis['units'] == '' or not isinstance(axis['units'], str):
                    axis['units'] = 'pixel'
                if not isinstance(axis['name'], str):
                    axis['name'] = str(key)

                dataset.set_dimension(int(key[-1]), sidpy.Dimension(dim_array, name=axis['name'], units=axis['units'],
                                                                    quantity=axis['name'],
                                                                    dimension_type=dimension_type))
        elif s.data.ndim == 4:
            dataset.data_type = 'IMAGE_4D'
            for key, axis in axes.items():
                if axis['navigate']:
                    dimension_type = 'spatial'
                else:
                    dimension_type = 'reciprocal'
                dim_array = np.arange(axis['size']) * axis['scale'] + axis['offset']
                if axis['units'] == '' or not isinstance(axis['units'], str):
                    axis['units'] = 'pixel'
                if not isinstance(axis['name'], str):
                    axis['name'] = str(key)
                dataset.set_dimension(int(key[-1]), sidpy.Dimension(dim_array, name=axis['name'], units=axis['units'],
                                                                    quantity=axis['name'],
                                                                    dimension_type=dimension_type))
    dataset.metadata = dict(s.metadata)
    dataset.original_metadata = dict(s.original_metadata)
    dataset.title = dataset.metadata['General']['title']
    if 'quantity' in dataset.metadata['Signal']:
        dataset.units = dataset.metadata['Signal']['quantity'].split('(')[-1][:-1]
        dataset.quantity = dataset.metadata['Signal']['quantity'].split('(')[0]
    dataset.source = 'hyperspy'
    return dataset



#!/usr/bin/env python
# -*- coding: iso-8859-1 -*-

################################################################################
# Python class for reading EDAX .h5 files into sidpy Dataset
# and extracting all metadata
#
# Written by Gerd Duscher, UTK 2023
#
# Works for python 3
#
################################################################################

import struct
import h5py
# from warnings import warn
import sys
import numpy as np
import os

import sidpy

__all__ = ["EDAXReader", "version"]

version = '0.1beta'

debugLevel = 0  # 0=none, 1-3=basic, 4-5=simple, 6-10 verbose

if sys.version_info.major == 3:
    unicode = str

# ### utility functions ###




def get_dataset_keys(h5_object):
    if not isinstance(h5_object, (h5py.Group, h5py.File)):
        raise TypeError('h5_object should be a h5py.File or h5py.Group object')

    dataset_list = []
    h5_object.visit(lambda key : dataset_list.append(key) if isinstance(h5_object[key], h5py.Dataset) else None)
    return dataset_list


def reader(h5_file):
    base_group = h5_file[list(h5_file.keys())[0]]
    all_datasets_list = get_dataset_keys(base_group)
    
    datasets = {}
    # read spectum  images first
    for dataset_item in all_datasets_list:
        print(dataset_item)
        if 'Live Map' in dataset_item:
            if 'SPD' in dataset_item:
                dataset = read_spectrum_image(base_group, dataset_item)

def read_metadata_array(structured_array):
    metadata = {}
    for i in range(len(structured_array.dtype.fields)):
        # print(structured_array.dtype.names[i], structured_array[structured_array.dtype.names[i]])
        metadata[structured_array.dtype.names[i]] = structured_array[structured_array.dtype.names[i]]
    return(metadata)

def read_spectrum_image(base_group, dataset_item):
    group_end = dataset_item.index('/SPD')
    dataset = sidpy.Dataset.from_array(base_group[dataset_item][:])
    if dataset.ndim !=3:
        print('Error reading spectrum image')
    dataset.original_metadata = read_metadata_array(base_group[dataset_item[:group_end]+'/SPC'][:])
    dataset.original_metadata.update(dict(base_group[dataset_item].attrs))
    # print(dataset.original_metadata['StartEnergy'], dataset.original_metadata['EndEnergy'])
    # print(dataset.shape[2]/dataset.original_metadata['EndEnergy'])
    # print(dataset.shape[2]*dataset.original_metadata['evPerChannel'] )
    # print(dataset.original_metadata['MicronPerPixelX'], dataset.original_metadata['MicronPerPixelY'])


    dataset.title = dataset_item[:group_end].replace('/', '_').replace(' ','')+'_SpectrumImage'
    dataset.data_type = 'spectral_image'
    dataset.set_dimension(0, sidpy.Dimension(np.arange(dataset.shape[0])* dataset.original_metadata['MicronPerPixelX']*1000,
                                                        name='x', units='nm',
                                                        quantity='distance',
                                                        dimension_type='spatial'))
    dataset.set_dimension(1, sidpy.Dimension(np.arange(dataset.shape[1])* dataset.original_metadata['MicronPerPixelY']*1000,
                                                          name='y', units='nm',
                                                          quantity='distance',
                                                          dimension_type='spatial'))
    dataset.set_dimension(2, sidpy.Dimension(np.arange(dataset.shape[2])* dataset.original_metadata['evPerChannel'],
                                                          name='energy_scale', units='eV',
                                                          quantity='energy',
                                                          dimension_type='spectral'))

    dataset.units = 'counts'
    dataset.quantity = 'intensity'
    
    # print(dataset.title)

    return dataset


def read_image(base_group, dataset_item):
    dataset = sidpy.Dataset.from_array(base_group[dataset_item][:])

    if dataset.ndim !=2:
        print('Error reading image')
    
    if dataset_item[-4:] == '.dat':
        dataset.original_metadata = read_metadata_array(base_group[dataset_item[:-3]+'ipr'][:])
        dataset.original_metadata.update(dict(base_group[dataset_item].attrs))
        dataset.title = dataset_item[:-4].replace('/', '_').replace(' ','')
    elif 'MAPIMAGE' in dataset_item:
        group_end = dataset_item.index('/MAPIMAGE')
        dataset.original_metadata = read_metadata_array(base_group[dataset_item[:group_end]+'/MAPIMAGEIPR'][:])
        dataset.original_metadata.update(read_metadata_array(base_group[dataset_item[:group_end]+'/MAPIMAGECOLLECTIONPARAMS'][:]))
        dataset.original_metadata.update(dict(base_group[dataset_item].attrs))
        dataset.title = dataset_item[:group_end].replace('/', '_').replace(' ','')+'_Image'
    elif 'FOVIMAGE' in dataset_item:
        group_end = dataset_item.index('/FOVIMAGE')
        dataset.original_metadata = read_metadata_array(base_group[dataset_item[:group_end]+'/FOVIPR'][:])
        dataset.original_metadata.update(read_metadata_array(base_group[dataset_item[:group_end]+'/FOVIMAGECOLLECTIONPARAMS'][:]))
        dataset.original_metadata.update(dict(base_group[dataset_item].attrs))
        dataset.title = dataset_item[:group_end].replace('/', '_').replace(' ','')+'_SurveyImage'

    dataset.data_type = 'image'
    dataset.units = 'counts'
    dataset.quantity = 'intensity'
    
    dataset.set_dimension(0, sidpy.Dimension(np.arange(dataset.shape[0])* dataset.original_metadata['MicronsPerPixelX']*1000,
                                                        name='x', units='nm',
                                                        quantity='distance',
                                                        dimension_type='spatial'))
    dataset.set_dimension(1, sidpy.Dimension(np.arange(dataset.shape[1])* dataset.original_metadata['MicronsPerPixelY']*1000,
                                                        name='y', units='nm',
                                                        quantity='distance',
                                                        dimension_type='spatial'))
    return dataset
           
class EDAXReader(sidpy.Reader):
    """
    Creates an instance of EDAXReader which can read one or more HDF5
    datasets formatted in the EDAX format

    We can read Images, and SpectrumStreams (SpectrumImages and Spectra).
    Please note that all original metadata are retained in each sidpy dataset.

    Parameters
    ----------
    file_path : str
        Path to a EDAX file
    Return
    ------
    datasets: dict
        dictionary of sidpy.Datasets
    """

    def __init__(self, file_path, verbose=False):
        """
        file_path: filepath to dm3 file.
        """

        super().__init__(file_path)

        # initialize variables ##
        self.verbose = verbose
        self.__filename = file_path

        path, file_name = os.path.split(self.__filename)
        self.basename, self.extension = os.path.splitext(file_name)
        self.datasets = {}
        
        if 'h5' in self.extension:
            try:
                h5_file = h5py.File(self.__filename, mode='r')
                if not isinstance(h5_file, h5py.File):
                    raise TypeError("File {} does not seem to be of EDAX`s .h5 format".format(self.__filename))
        
                if len(h5_file.keys()) != 1:
                    raise TypeError("File {} does not seem to be of EDAX`s .h5 format".format(self.__filename))
                
                base_metadata = dict(h5_file[list(h5_file.keys())[0]].attrs)
                edax_file = False
                if 'Company' in base_metadata:
                    if base_metadata['Company'] == 'EDAX, LLC':
                        edax_file =True
                if not edax_file:
                    raise TypeError("File {} does not seem to be of EDAX`s .h5 format".format(self.__filename))
                h5_file.close()
            except IOError:
                raise IOError("File {} does not seem to be of EDAX's .h5 format".format(self.__filename))

    def read(self):
        if 'h5' in self.extension:
            # TODO: use lazy load for large datasets
            self.__f = h5py.File(self.__filename, 'r')
            base_group = self.__f[list(self.__f.keys())[0]]
            all_datasets_list = get_dataset_keys(base_group)
    
            for dataset_item in all_datasets_list:
                if 'Live Map' in dataset_item:
                    if 'SPD' in dataset_item:
                        dataset = read_spectrum_image(base_group, dataset_item)
                        self.datasets[dataset.title] = dataset
                        group_end = dataset_item.index('/SPD')
                        self.extract_crucial_metadata(dataset.title)
                        if dataset_item[:group_end]+'/MAPIMAGE' in all_datasets_list:
                            image_dataset = read_image(base_group, dataset_item[:group_end]+'/MAPIMAGE')
                            self.datasets[image_dataset.title] = image_dataset
                            self.extract_crucial_metadata(image_dataset.title)
                        parent_group = dataset_item.index('/Live Map')
                        if dataset_item[:parent_group]+'/FOVIMAGE' in all_datasets_list:
                            image_dataset = read_image(base_group,dataset_item[:parent_group]+'/FOVIMAGE')
                            self.datasets[image_dataset.title] = image_dataset
                            self.extract_crucial_metadata(image_dataset.title)
                        for item in all_datasets_list:    
                            if dataset_item[:-3]+'ROIs' in item:
                                if item[-3:] == 'dat':
                                    image_dataset = read_image(base_group,item)
                                    self.datasets[image_dataset.title] = image_dataset
                                    self.extract_crucial_metadata(image_dataset.title)

            self.__f.close()
            return self.datasets

    def get_filename(self):
        return self.__filename

    filename = property(get_filename)


    def get_tags(self):
        return self.original_metadata

    tags = property(get_tags)

    def get_datasets(self):
        return self.datasets 

    def extract_crucial_metadata(self, key):
        """Read essential parameter from original_metadata originating from a Nion file"""
        
        original_metadata = self.datasets[key].original_metadata

        if not isinstance(original_metadata, dict):
            raise TypeError('We need a dictionary to read')
        
        experiment = {'stage': {},
                      'detector': {},
                      'microscope': {},
                      'analysis': {}}

        if 'KVolt' in original_metadata:
            experiment['microscope']['acceleration_voltage_V'] = original_metadata['KVolt']*100
        if 'KV' in original_metadata:
            experiment['microscope']['acceleration_voltage_V'] = original_metadata['KV']*1000
        if 'WorkingDistance' in original_metadata:
            experiment['stage']['working_distance'] = original_metadata['WorkingDistance']/1000
        if 'LiveTime' in original_metadata:
            experiment['detector']['live_time'] = original_metadata['LiveTime']
        if 'TiltAngle' in original_metadata:
            experiment['detector']['tilt_angle'] = original_metadata['TiltAngle']
        if 'TakeOffAngle' in original_metadata:
            experiment['detector']['take_off_angle'] = original_metadata['TakeOffAngle']
        if 'DetectorResoultion' in original_metadata:
            experiment['detector']['resolution'] = original_metadata['DetectorResoultion']
        if 'AlThickness' in original_metadata:
            experiment['detector']['Al_thickness'] = original_metadata['AlThickness'] * 1e-6 # in m
        if 'BeWinThickness' in original_metadata:
            experiment['detector']['Be_thickness'] = original_metadata['BeWinThickness'] * 1e-6 # in m
        if 'ParThickness' in original_metadata:
            experiment['detector']['Par_thickness'] = original_metadata['ParThickness'] * 1e-6 # in m
        if 'AuThickness' in original_metadata:
            experiment['detector']['Au_thickness'] = original_metadata['AuThickness'] * 1e-6 # in m
        if 'SiDeadThickness' in original_metadata:
            experiment['detector']['Si_dead_layer_thickness'] = original_metadata['SiDeadThickness'] * 1e-6 # in m
        if 'SiLiveThickness' in original_metadata:
            experiment['detector']['Si_live_thickness'] = original_metadata['SiLiveThickness'] # in m
            
        if 'XRayIncidenceAngle' in original_metadata:
            experiment['detector']['x_ray_incident_angle'] = original_metadata['XRayIncidenceAngle']
        if 'AzimuthAngle' in original_metadata:
            experiment['detector']['azimuth_angle'] = original_metadata['AzimuthAngle']
        if 'ElevationAngle' in original_metadata:
            experiment['detector']['elevation_angle'] = original_metadata['ElevationAngle']
        if 'BCoefficient' in original_metadata:
            experiment['detector']['b_coefficient'] = original_metadata['BCoefficient']
        if 'CCoefficient' in original_metadata:
            experiment['detector']['c_coefficient'] = original_metadata['CCoefficient']

        if 'XTiltAngle' in original_metadata:
            experiment['stage']['x_tilt_angle'] = original_metadata['XTiltAngle']
        if 'YTiltAngle' in original_metadata:
            experiment['stage']['y_tilt_angle'] = original_metadata['YTiltAngle']
        if 'Tilt' in original_metadata:
            experiment['stage']['tilt'] = original_metadata['Tilt']/10
        
        if 'AtomicNumberOfPeakIds' in original_metadata:
            experiment['analysis']['atomic_numbers'] = original_metadata['AtomicNumberOfPeakIds']
        if 'EnergyOfPeakIds' in original_metadata:
            experiment['analysis']['peak_energies'] = original_metadata['EnergyOfPeakIds']
        if 'WeightFraction'in original_metadata:
            experiment['analysis']['weight_fractions'] = original_metadata['WeightFraction']
        
        self.datasets[key].metadata['experiment'] = experiment





"""
Utilities for automated ingestion of data and metadata from proprietary file formats
"""

from warnings import warn
from .readers import all_readers


def ingest(file_path):
    """
    Extracts sidpy.Dataset objects from proprietary data files.
    The sidpy.Dataset object(s) would contain both raw data and metadata.
    More than one Dataset object could be returned from this function.

    Parameters
    ----------
    file_path : str
        Path to raw data file

    Returns
    -------
    list
        List of sidpy.Dataset objects
    """
    valid_readers = []
    for this_reader in all_readers:
        try:
            t = this_reader(file_path)
        except Exception:
            continue
        try:
            # TODO: This code should be absorbed into the constructor. That can fail anyway
            readable = t.can_read()
        except Exception:
            continue
        if readable:
            valid_readers.append(this_reader)
            # print('{} is a valid reader'.format(this_reader))
            
    final_reader = valid_readers[0]
    if len(valid_readers) == 0:
        raise TypeError(
            "The automatic search for a suitable Reader was unsuccessful")
    elif len(valid_readers) > 1:
        warn("Multiple Reader s may be able to read your file."
             + "The {} will be applied.".format(valid_readers[-1])
             + " Consider specifying 'force_translator'")
        final_reader = valid_readers[-1]
    
    try:
        extracted = final_reader(file_path).read()
    except Exception:
        return None

    return extracted



"""
This file is taken directly from naononispy, which is at https://github.com/underchemist/nanonispy
and written by @underchemist

Due to a change to numpy, the existing version of nanonispy does not build. 

The fixed version appears below and is also given as a pull request  to the original package.

Note that nanonispy was released with the MIT license, as is SciFiReaders. 

Read details about the license in the LICENSE file.

"""

import os
import warnings
import numpy as np

nanonis_format_dict = {
    'big endian float 32': '>f4',
    'little endian float 32': '<f4',
    'big endian float 64': '>f8',
    'little endian float 64': '<f8',
}

nanonis_end_tags = dict(grid=':HEADER_END:', scan='SCANIT_END', spec='[DATA]')

class NanonisFile:

    """
    Base class for Nanonis data files (grid, scan, point spectroscopy).

    Handles methods and parsing tasks common to all Nanonis files.

    Parameters
    ----------
    fname : str
        Name of Nanonis file.

    Attributes
    ----------
    datadir : str
        Directory path for Nanonis file.
    basename : str
        Just the filename, no path.
    fname : str
        Full path of Nanonis file.
    filetype : str
        filetype corresponding to filename extension.
    byte_offset : int
        Size of header in bytes.
    header_raw : str
        Unproccessed header information.
    """

    def __init__(self, fname):
        _data_format = nanonis_format_dict
        self.datadir, self.basename = os.path.split(fname)
        self.fname = fname
        self.filetype = self._determine_filetype()
        self.byte_offset = self.start_byte()
        self.header_raw = self.read_raw_header(self.byte_offset)

    def _determine_filetype(self):
        """
        Check last three characters for appropriate file extension,
        raise error if not.

        Returns
        -------
        str
            Filetype name associated with extension.

        Raises
        ------
        UnhandledFileError
            If last three characters of filename are not one of '3ds',
            'sxm', or 'dat'.
        """

        _, fname_ext = os.path.splitext(self.fname)
        if fname_ext == '.3ds':
            return 'grid'
        elif fname_ext == '.sxm':
            return 'scan'
        elif fname_ext == '.dat':
            return 'spec'
        else:
            raise UnhandledFileError('{} is not a supported filetype or does not exist'.format(self.basename))

    def read_raw_header(self, byte_offset):
        """
        Return header as a raw string.

        Everything before the end tag is considered to be part of the header.
        the parsing will be done later by subclass methods.

        Parameters
        ----------
        byte_offset : int
            Size of header in bytes. Read up to this point in file.

        Returns
        -------
        str
            Contents of filename up to byte_offset as a decoded binary
            string.
        """

        with open(self.fname, 'rb') as f:
            return f.read(byte_offset).decode('utf-8', errors='replace')

    def start_byte(self):
        """
        Find first byte after end tag signalling end of header info.

        Caveat, I believe this is the first byte after the end of the
        line that the end tag is found on, not strictly the first byte
        directly after the end tag is found. For example in Scan
        __init__, byte_offset is incremented by 4 to account for a
        'start' byte that is not actual data.

        Returns
        -------
        int
            Size of header in bytes.
        """

        with open(self.fname, 'rb') as f:
            tag = nanonis_end_tags[self.filetype]

            # Set to a default value to know if end_tag wasn't found
            byte_offset = -1

            for line in f:
                # Convert from bytes to str
                try:
                    entry = line.strip().decode()
                except UnicodeDecodeError:
                    warnings.warn('{} has non-uft-8 characters, replacing them.'.format(f.name))
                    entry = line.strip().decode('utf-8', errors='replace')
                if tag in entry:
                    byte_offset = f.tell()
                    break

            if byte_offset == -1:
                raise FileHeaderNotFoundError(
                        'Could not find the {} end tag in {}'.format(tag, self.basename)
                        )

        return byte_offset

    def set_data_format(self, data_format):
        # default value is '>f4' big endian float 32 bit
        if data_format is None:
            self.data_format = nanonis_format_dict['big endian float 32']
        else:
            try:
                self.data_format = nanonis_format_dict[data_format]
            except KeyError as exc:
                self.data_format = nanonis_format_dict['big endian float 32']
                warnings.warn('{} is not a valid data format'.format(data_format))

class Grid(NanonisFile):

    """
    Nanonis grid file class.

    Contains data loading method specific to Nanonis grid file. Nanonis
    3ds files contain a header terminated by '\r\n:HEADER_END:\r\n'
    line, after which big endian encoded binary data starts. A grid is
    always recorded in an 'up' direction, and data is recorded
    sequentially starting from the first pixel. The number of bytes
    corresponding to a single pixel will depend on the experiment
    parameters. In general the size of one pixel will be a sum of

        - # fixed parameters
        - # experimental parameters
        - # sweep signal points (typically bias).

    Hence if there are 2 fixed parameters, 8 experimental parameters,
    and a 512 point bias sweep, a pixel will account 4 x (522) = 2088
    bytes of data. The class intuits this from header info and extracts
    the data for you and cuts it up into each channel, though normally
    this should be just the current.

    Currently cannot accept grids that are incomplete.

    Parameters
    ----------
    fname : str
        Filename for grid file.
    header_override : dict, optional
        A dict of key:value to override any corresponding key:value should
        they be wrong or missing in your header. Keys in header_override must
        match keys in Grid.header_raw.

    Attributes
    ----------
    header : dict
        Parsed 3ds header. Relevant fields are converted to float,
        otherwise most are string values.
    signals : dict
        Dict keys correspond to channel name, with values being the
        corresponding data array.

    Raises
    ------
    UnhandledFileError
        If fname does not have a '.3ds' extension.
    """

    def __init__(self, fname, header_override=None, data_format=None):
        _is_valid_file(fname, ext='3ds')
        super().__init__(fname)
        self.set_data_format(data_format)
        self.header = _parse_3ds_header(self.header_raw, header_override=header_override)
        self.signals = self._load_data()
        self.signals['sweep_signal'] = self._derive_sweep_signal()
        self.signals['topo'] = self._extract_topo()

    def _load_data(self):
        """
        Read binary data for Nanonis 3ds file.

        Returns
        -------
        dict
            Channel name keyed dict of 3d array.
        """
        # load grid params
        nx, ny = self.header['dim_px']
        num_sweep = self.header['num_sweep_signal']
        num_param = self.header['num_parameters']
        num_chan = self.header['num_channels']
        data_dict = dict()

        # open and seek to start of data
        f = open(self.fname, 'rb')
        f.seek(self.byte_offset)
        data_format = self.data_format
        griddata = np.fromfile(f, dtype=data_format)
        f.close()

        # pixel size in bytes
        exp_size_per_pix = num_param + num_sweep*num_chan

        # resize from 1d to 3d
        griddata.resize((ny, nx, exp_size_per_pix))

        # experimental parameters are first num_param of every pixel
        params = griddata[:, :, :num_param]
        data_dict['params'] = params

        # extract data for each channel
        for i, chann in enumerate(self.header['channels']):
            start_ind = num_param + i * num_sweep
            stop_ind = num_param + (i+1) * num_sweep
            data_dict[chann] = griddata[:, :, start_ind:stop_ind]

        return data_dict

    def _derive_sweep_signal(self):
        """
        Computer sweep signal.

        Based on start and stop points of sweep signal in header, and
        number of sweep signal points.

        Returns
        -------
        numpy.ndarray
            1d sweep signal, should be sample bias in most cases.
        """
        # find sweep signal start and end from a given pixel value
        sweep_start, sweep_end = self.signals['params'][0, 0, :2]
        num_sweep_signal = self.header['num_sweep_signal']

        return np.linspace(sweep_start, sweep_end, num_sweep_signal, dtype=float)

    def _extract_topo(self):
        """
        Extract topographic map based on z-controller height at each
        pixel.

        The data is already extracted, though it lives in the signals
        dict under the key 'parameters'. Currently the 4th column is the
        Z (m) information at each pixel, should update this to be more
        general in case the fixed/experimental parameters are not the
        same for other Nanonis users.

        Returns
        -------
        numpy.ndarray
            Copy of already extracted data to be more easily accessible
            in signals dict.
        """
        return self.signals['params'][:, :, 4]


class Scan(NanonisFile):

    """
    Nanonis scan file class.

    Contains data loading methods specific to Nanonis sxm files. The
    header is terminated by a 'SCANIT_END' tag followed by the \1A\04
    code. The NanonisFile header parse method doesn't account for this
    so the Scan __init__ method just adds 4 bytes to the byte_offset
    attribute so as to not include this as a datapoint.

    Data is structured a little differently from grid files, obviously.
    For each pixel in the scan, each channel is recorded forwards and
    backwards one after the other.

    Currently cannot take scans that do not have both directions
    recorded for each channel, nor incomplete scans.

    Parameters
    ----------
    fname : str
        Filename for scan file.

    Attributes
    ----------
    header : dict
        Parsed sxm header. Some fields are converted to float,
        otherwise most are string values.
    signals : dict
        Dict keys correspond to channel name, values correspond to
        another dict whose keys are simply forward and backward arrays
        for the scan image.

    Raises
    ------
    UnhandledFileError
        If fname does not have a '.sxm' extension.
    """

    def __init__(self, fname, data_format=None):
        _is_valid_file(fname, ext='sxm')
        super().__init__(fname)
        self.set_data_format(data_format)
        self.header = _parse_sxm_header(self.header_raw)

        # data begins with 4 byte code, add 4 bytes to offset instead
        self.byte_offset += 4

        # load data
        self.signals = self._load_data()


    def _load_data(self):
        """
        Read binary data for Nanonis sxm file.

        Returns
        -------
        dict
            Channel name keyed dict of each channel array.
        """
        channs = list(self.header['data_info']['Name'])
        nchanns = len(channs)
        nx, ny = self.header['scan_pixels']

        # assume both directions for now
        ndir = 2

        data_dict = dict()

        # open and seek to start of data
        f = open(self.fname, 'rb')
        f.seek(self.byte_offset)
        data_format = self.data_format
        scandata = np.fromfile(f, dtype=data_format)
        f.close()

        # reshape
        scandata_shaped = scandata.reshape(nchanns, ndir, ny, nx)

        # extract data for each channel
        for i, chann in enumerate(channs):
            chann_dict = dict(forward=scandata_shaped[i, 0, :, :],
                              backward=scandata_shaped[i, 1, :, :])
            data_dict[chann] = chann_dict

        return data_dict


class Spec(NanonisFile):

    """
    Nanonis point spectroscopy file class.

    These files are a little easier to handle since they are stored in
    ascii format.

    Parameters
    ----------
    fname : str
        Filename for spec file.

    Attributes
    ----------
    header : dict
        Parsed dat header.

    Raises
    ------
    UnhandledFileError
        If fname does not have a '.dat' extension.
    """

    def __init__(self, fname):
        _is_valid_file(fname, ext='dat')
        super().__init__(fname)
        self.header = _parse_dat_header(self.header_raw)
        self.signals = self._load_data()

    def _load_data(self):
        """
        Loads ascii formatted .dat file.

        Header ended by '[DATA]' tag.

        Returns
        -------
        dict
            Keys correspond to each channel recorded, including
            saved/filtered versions of other channels.
        """

        # done differently since data is ascii, not binary
        f = open(self.fname, 'r')
        f.seek(self.byte_offset)
        data_dict = dict()

        column_names = f.readline().strip('\n').split('\t')
        f.close()
        num_lines = self._num_header_lines()
        specdata = np.genfromtxt(self.fname, delimiter='\t', skip_header=num_lines)

        for i, name in enumerate(column_names):
            data_dict[name] = specdata[:, i]

        return data_dict

    def _num_header_lines(self):
        """Number of lines the header is composed of"""
        with open(self.fname, 'r') as f:
            data = f.readlines()
            for i, line in enumerate(data):
                if nanonis_end_tags['spec'] in line:
                    return i + 2  # add 2 to skip the tag itself and column names
        return 0


class UnhandledFileError(Exception):

    """
    To be raised when unknown file extension is passed.
    """
    pass


class FileHeaderNotFoundError(Exception):

    """
    To be raised when no header information could be determined.
    """
    pass


def _parse_3ds_header(header_raw, header_override):
    """
    Parse raw header string.

    Empirically done based on Nanonis header structure. See Grid
    docstring or Nanonis help documentation for more details.

    Parameters
    ----------
    header_raw : str
        Raw header string from read_raw_header() method.

    Returns
    -------
    dict
        Channel name keyed dict of 3d array.
    """
    # cleanup string and remove end tag as entry
    header_entries = header_raw.split('\r\n')
    header_entries = header_entries[:-2]

    # Convert the strings to a dictionary.
    raw_dict = dict()
    for entry in header_entries:
        key, val = _split_header_entry(entry)
        raw_dict[key] = val

    if header_override is not None:
        for key, val in header_override.items():
            raw_dict[key] = val  # creates new entry if key doesn't match key in raw_dict

    # Transfer parameters from raw_dict to header_dict
    # Get the expected parameters first
    header_dict = dict()

    try:
        # grid dimensions in pixels
        header_dict['dim_px'] = [int(val) for val in raw_dict['Grid dim'].split(' x ')]
        raw_dict.pop('Grid dim')

        # grid frame center position, size, angle. Assumes len(raw_dict['Grid settings']) = 4
        header_dict['pos_xy'] = [float(val) for val in raw_dict['Grid settings'][:2]]
        header_dict['size_xy'] = [float(val) for val in raw_dict['Grid settings'][2:4]]
        header_dict['angle'] = float(raw_dict['Grid settings'][4])
        raw_dict.pop('Grid settings')

        # sweep signal
        header_dict['sweep_signal'] = raw_dict['Sweep Signal']
        raw_dict.pop('Sweep Signal')

        # fixed parameters
        header_dict['fixed_parameters'] = raw_dict['Fixed parameters']
        raw_dict.pop('Fixed parameters')

        # experimental parameters
        header_dict['experimental_parameters'] = raw_dict['Experiment parameters']
        raw_dict.pop('Experiment parameters')

        # number of parameters (each 4 bytes)
        header_dict['num_parameters'] = int(raw_dict['# Parameters (4 byte)'])
        raw_dict.pop('# Parameters (4 byte)')

        # experiment size in bytes
        header_dict['experiment_size'] = int(raw_dict['Experiment size (bytes)'])
        raw_dict.pop('Experiment size (bytes)')

        # number of points of sweep signal
        header_dict['num_sweep_signal'] = int(raw_dict['Points'])
        raw_dict.pop('Points')

        # channel names
        header_dict['channels'] = raw_dict['Channels']
        if type(header_dict['channels']) == str:
            # will be str if only one channel, make list of str so number of channels can be counted properly
            l = []
            l.append(header_dict['channels'])
            header_dict['channels'] = l
        header_dict['num_channels'] = len(header_dict['channels'])
        raw_dict.pop('Channels')

        # measure delay
        header_dict['measure_delay'] = float(raw_dict['Delay before measuring (s)'])
        raw_dict.pop('Delay before measuring (s)')

        # metadata
        header_dict['experiment_name'] = raw_dict['Experiment']
        header_dict['start_time'] = raw_dict['Start time']
        header_dict['end_time'] = raw_dict['End time']
        header_dict['user'] = raw_dict['User']
        header_dict['comment'] = raw_dict['Comment']
        raw_dict.pop('Experiment')
        raw_dict.pop('Start time')
        raw_dict.pop('End time')
        raw_dict.pop('User')
        raw_dict.pop('Comment')

    except (KeyError, ValueError) as e:
        msg = ' You can edit your header file or provide an override value in header_override'

        # guide user to using override dict
        if isinstance(e, KeyError):
            raise KeyError('[{key}] is missing from header.'.format(key=e.args[0]) + msg)
        elif isinstance(e, ValueError):
            print(e.args)
            raise ValueError('Unexpected value found in header.' + msg)
        else:
            raise

    # fold remaining header entries into dict
    for key, val in raw_dict.items():
        header_dict[key] = val

    return header_dict


def _parse_sxm_header(header_raw):
    """
    Parse raw header string.

    Empirically done based on Nanonis header structure. See Scan
    docstring or Nanonis help documentation for more details.

    Parameters
    ----------
    header_raw : str
        Raw header string from read_raw_header() method.

    Returns
    -------
    dict
        Channel name keyed dict of each channel array.
    """
    header_entries = header_raw.split('\n')
    header_entries = header_entries[:-3]

    header_dict = dict()
    entries_to_be_split = ['scan_offset',
                           'scan_pixels',
                           'scan_range',
                           'scan_time']

    entries_to_be_floated = ['scan_offset',
                             'scan_range',
                             'scan_time',
                             'bias',
                             'acq_time']

    entries_to_be_inted = ['scan_pixels']

    entries_to_be_dict = [':DATA_INFO:',
                          ':Z-CONTROLLER:',
                          ':Multipass-Config:']

    entries_possibly_multilines = [':COMMENT:']

    for i, entry in enumerate(header_entries):
        if entry in entries_to_be_dict:
            count = 1
            for j in range(i+1, len(header_entries)):
                if header_entries[j].startswith(':'):
                    break
                if header_entries[j][0] == '\t':
                    count += 1
            header_dict[entry.strip(':').lower()] = _parse_scan_header_table(header_entries[i+1:i+count])
            continue
        if entry in entries_possibly_multilines:
            multilines_entry = []
            for j in range(i+1, len(header_entries)):
                if header_entries[j].startswith(':'):
                    break
                multilines_entry.append(header_entries[j])
            header_dict[entry.strip(':').lower()] = '\n'.join(multilines_entry)
            continue
        if entry.startswith(':'):
            header_dict[entry.strip(':').lower()] = header_entries[i+1].strip()

    for key in entries_to_be_split:
        header_dict[key] = header_dict[key].split()

    for key in entries_to_be_floated:
        if isinstance(header_dict[key], list):
            header_dict[key] = np.asarray(header_dict[key], dtype=float)
        else:
            header_dict[key] = float(header_dict[key])
    for key in entries_to_be_inted:
        header_dict[key] = np.asarray(header_dict[key], dtype=int)

    return header_dict


def _parse_dat_header(header_raw):
    """
    Parse point spectroscopy header.

    Each key-value pair is separated by '\t' characters. Values may be
    further delimited by more '\t' characters.

    Returns
    -------
    dict
        Parsed point spectroscopy header.
    """
    header_entries = header_raw.split('\r\n')
    header_entries = header_entries[:-3]
    header_dict = dict()
    for entry in header_entries:
        # homogenize output of .dat files with \t delimit at end of every key
        if entry[-1] == '\t':
            entry = entry[:-1]
        if '\t' not in entry:
            entry += '\t'

        key, val = entry.split('\t')
        header_dict[key] = val

    return header_dict


def _clean_sxm_header(header_dict):
    """
    Cleanup header dicitonary key-value pairs.

    Parameters
    ----------
    header_dict : dict
        Should be dict returned from _parse_sxm_header method.

    Returns
    -------
    clean_header_dict : dict
        Cleaned header dictionary.
    """
    pass


def _split_header_entry(entry):
    """
    Split 3ds header entries by '=' character. If multiple values split
    those by ';' character.
    """

    key_str, val_str = entry.split("=", 1)

    if ';' in val_str:
        return key_str, (val_str.strip('"').split(';'))
    else:
        return key_str, val_str.strip('"')


def save_array(file, arr, allow_pickle=True):
    """
    Wrapper to numpy.save method for arrays.

    The idea would be to use this to save a processed array for later
    use in a matplotlib figure generation scripts. See numpy.save
    documentation for details.

    Parameters
    ----------
    file : file or str
        File or filename to which the data is saved.  If file is a file-
        object, then the filename is unchanged.  If file is a string, a
        ``.npy`` extension will be appended to the file name if it does
        not already have one.
    arr : array_like
        Array data to be saved.
    allow_pickle : bool, optional
        Allow saving object arrays using Python pickles. Reasons for
        disallowing pickles include security (loading pickled data can
        execute arbitrary code) and portability (pickled objects may not
        be loadable on different Python installations, for example if
        the stored objects require libraries that are not available, and
        not all pickled data is compatible between Python 2 and Python
        3). Default: True
    """
    np.save(file, arr, allow_pickle=allow_pickle)


def load_array(file, allow_pickle=True):
    """
    Wrapper to numpy.load method for binary files.

    See numpy.load documentation for more details.

    Parameters
    ----------
    file : file or str
        The file to read. File-like objects must support the
    ``seek()`` and ``read()`` methods. Pickled files require that the
    file-like object support the ``readline()`` method as well.
    allow_pickle : bool, optional
        Allow loading pickled object arrays stored in npy files. Reasons
        for disallowing pickles include security, as loading pickled
        data can execute arbitrary code. If pickles are disallowed,
        loading object arrays will fail. Default: True

    Returns
    -------
    result : array, tuple, dict, etc.
        Data stored in the file. For ``.npz`` files, the returned
        instance of NpzFile class must be closed to avoid leaking file
        descriptors.
    """
    return np.load(file)


def _parse_scan_header_table(table_list):
    """
    Parse scan file header entries whose values are tab-separated
    tables.
    """
    table_processed = []
    for row in table_list:
        # strip leading \t, split by \t
        table_processed.append(row.strip('\t').split('\t'))

    # column names are first row
    keys = table_processed[0]
    values = table_processed[1:]

    zip_vals = zip(*values)

    return dict(zip(keys, zip_vals))


def _is_valid_file(fname, ext):
    """
    Detect if invalid file is being initialized by class.
    """
    _, fname_ext = os.path.splitext(fname)
    if fname_ext[1:] != ext:
        raise UnhandledFileError('{} is not a {} file'.format(fname, ext))


"""
Tools to read, data in from TEM files

Submodules
----------

.. autosummary::
    :toctree: _autosummary

    dm_reader
    nion_reader
"""
from .dm_reader import DMReader, DM3Reader
from .nion_reader import NionReader
from .emd_reader import EMDReader
from .mrc_reader import MRCReader

__all__ = ['DMReader', 'DM3Reader', 'NionReader', 'EMDReader', 'MRCReader']

all_readers = [DMReader, DM3Reader, NionReader, EMDReader, MRCReader]


# lines (164 sloc)  7.4 KB

"""
:class:`~ScopeReaders.generic.image.ImageTranslator` class that extracts the
data from typical image files into :class:`~sidpy.Dataset` objects
Created on Feb 9, 2016
@author: Mani Valleti
"""

import os
import sys

import numpy as np
import PIL
from sidpy import Dataset, Dimension, Reader

try:
    import tifffile as tff
except ModuleNotFoundError:
    tff = None


class ImageReader(Reader):
    """
    Translates data from an image file to a sidpy dataset
    """

    def __init__(self, file_path, *args, **kwargs):
        super().__init__(file_path, *args, **kwargs)
        image_path = self._parse_file_path(self._input_file_path)
        ext = os.path.splitext(image_path)[-1]
        if ext not in ['.jpg', '.jpeg', '.png', '.bmp', '.tif', '.tiff']:
            raise NotImplementedError(
                'The provided file type {} is not supported by the reader as of now \n'.format(ext)
                + 'Please provide one of "jpg", "jpeg", "png", "bmp", ".tif", ".tiff" file types')

    @staticmethod
    def _parse_file_path(image_path):
        """
        Returns a list of all files in the directory given by path
        Parameters
        ---------------
        image_path : str
            absolute path to the image file
        Returns
        ----------
        image_path : str
            Absolute file path to the image
        """
        if not isinstance(image_path, str):
            raise TypeError("'image_path' argument for ImageReader should be a str")
        if not os.path.exists(os.path.abspath(image_path)):
            raise FileNotFoundError('Specified image does not exist.')
        else:
            image_path = os.path.abspath(image_path)

        return image_path

    def read(self, **image_args):
        """
        Translates the image in the provided file into a sidpy Dataset
        Parameters
        ----------------
        image_args : dict
            Arguments to be passed to read_image.  Arguments depend on the type of image.
        Returns
        ----------
        sidpy Dataset
        """

        image_path = self._parse_file_path(self._input_file_path)

        img_data, dimensions, metadata, original_metadata = read_image(image_path, **image_args)

        # Assuming that we have already dealt with classifying the shape
        # We are only as good as the tifffile package here

        # Working around occasional "cannot modify read-only array" error
        image = img_data.copy()

        data_set = Dataset.from_array(image, title='Image')

        data_set.data_type = 'image'
        # Can we somehow get the units and/or the quantity from the metadata??
        data_set.units = 'a. u.'
        data_set.quantity = 'Intensity'
        data_set.metadata = metadata.copy()
        data_set.original_metadata = original_metadata.copy()

        for i, dim in enumerate(dimensions):
            data_set.set_dimension(i, dim.copy())

        return data_set

    def can_read(self):
        """
        Tests whether or not the provided file has the appropriate extensions
        Returns
        -------

        """
        exts = ['jpg', 'jpeg', 'png', 'tiff', 'bmp', 'csv', 'txt']
        return super(ImageReader, self).can_read(extension=exts)


def read_image(image_path, *args, **kwargs):
    """
    Read the image file at `image_path` into a numpy array either via numpy (.txt)
    or via tiffifle(.tif) or via pillow (.jpg, etc.)

    Parameters
    ----------
    image_path : str
        Path to the image file

    Returns
    -------
    image : :class:`numpy.ndarray`
        Array containing the image from the file `image_path`.
    """

    # As of now we are not trying to read any metadata from these image formats.
    # We will cross that bridge when someone raises an issue
    ext = os.path.splitext(image_path)[-1]
    original_metadata, metadata, dimensions = {}, {}, []
    if ext in ['.jpg', '.jpeg', '.png', '.bmp']:
        img_data = np.asarray(PIL.Image.open(image_path))
        # Colored images, color channel last
        # Here we assume that the file path provided has a single image and not a stack
        # For stacks we turn our attention to tiff
        if len(img_data.shape) == 3:
            dimensions.append(Dimension(np.arange(img_data.shape[0]), 'y',
                                        units='generic', quantity='Length',
                                        dimension_type='spatial'))

            dimensions.append(Dimension(np.arange(img_data.shape[1]), 'x',
                                        units='generic', quantity='Length',
                                        dimension_type='spatial'))

            dimensions.append(Dimension(np.arange(img_data.shape[2]), 's',
                                        units='generic', quantity='SamplesPerPixel',
                                        dimension_type='frame'))
        # Grayscale single images
        if len(img_data.shape) == 2:
            dimensions.append(Dimension(np.arange(img_data.shape[0]), 'y',
                                        units='generic', quantity='Length',
                                        dimension_type='spatial'))

            dimensions.append(Dimension(np.arange(img_data.shape[1]), 'x',
                                        units='generic', quantity='Length',
                                        dimension_type='spatial'))

        return img_data, dimensions, metadata, original_metadata

    elif ext in ['.tif', '.tiff']:
        if tff is None:
            raise ModuleNotFoundError("tifffile is not installed")
        else:
            tif = tff.TiffFile(image_path)
            img_data = tif.asarray()

        # Only single series is supported for now, and for definition of a series refer the tifffile package
        img_shape, img_axes = tif.series[0].shape, tif.series[0].axes

        # Dealing with metadata that's common to the whole image, we will place this in original_metadata
        # First let's filter out all the attributes that end with metadata
        metadata_names = [a for a in dir(tif) if (not a.startswith('__') or not a.startswith('_'))
                          and a.endswith('_metadata')]
        for name in metadata_names:
            if getattr(tif, name) is not None:
                original_metadata[name] = getattr(tif, name)

        # Now metadata corresponding to individual frames. This is placed in metadata
        for page in tif.pages:
            for tag in page.tags:
                metadata[tag.name] = tag.value

        # Dealing with axes and dimensions
        # We will use the following dictionary to map tifffile axes to sidpy dimensions
        axes_dict = {'X': ['x', 'Length', 'spatial'],
                     'Y': ['y', 'Width', 'spatial'],
                     'Z': ['z', 'Depth', 'spatial'],
                     'S': ['s', 'SamplesPerPixel', 'frame'],
                     'T': ['t', 'Time', 'time'],
                     'C': ['c', 'Channel', 'frame'],
                     'Q': ['q', 'other', 'UNKNOWN']
                     }

        for i in range(len(img_shape)):
            if img_axes[i] == 'X' or img_axes[i] == 'Y':
                # We are here to handle X and Y dimensions
                # Here it is assumed that all the channels/time frames/depth frames have the same resolution unit

                res_name = 'XResolution' if img_axes[i] == 'X' else 'YResolution'
                for page in tif.pages:
                    if res_name in page.tags and 'ResolutionUnit' in page.tags:
                        if page.tags['ResolutionUnit'].value.value == 1:
                            # No unit provided
                            if isinstance(page.tags[res_name].value, tuple):
                                if page.tags[res_name].value == (1, 1):
                                    res, unit = 1, 'generic'
                                else:
                                    'It is assumed that the resolutions is given in pixels per inch'
                                    res, unit = page.tags[res_name].value[0], 'inches'
                            else:
                                if page.tags[res_name].value == 1:
                                    res, unit = 1, 'generic'
                                else:
                                    unit, res = 'inches', page.tags[res_name].value

                        elif page.tags['ResolutionUnit'].value.value == 2:
                            unit = 'inches'
                            if isinstance(page.tags[res_name].value, tuple):
                                res = page.tags[res_name].value[0]
                            else:
                                # At this point it is assumed that resolution is not a tuple but just and int/float
                                res = page.tags[res_name].value

                        elif page.tags['ResolutionUnit'].value.value == 3:
                            unit = 'cms'
                            if isinstance(page.tags[res_name].value, tuple):
                                res = page.tags[res_name].value[1]
                            else:
                                # At this point it is assumed that resolution is not a tuple but just and int/float
                                res = page.tags[res_name].value
                        break
                dimensions.append(Dimension(np.arange(img_shape[i]) * (1. / res), axes_dict[img_axes[i]][0],
                                            quantity=axes_dict[img_axes[i]][1], units=unit,
                                            dimension_type=axes_dict[img_axes[i]][2]))

            else:
                dimensions.append(Dimension(np.arange(img_shape[i]), axes_dict[img_axes[i]][0],
                                            quantity=axes_dict[img_axes[i]][1],
                                            dimension_type=axes_dict[img_axes[i]][2]))

        return img_data, dimensions, metadata, original_metadata
